import datetime as dt
from http import HTTPStatus
from typing import List, NamedTuple, Set
from unittest.mock import patch

import pook

from django.utils.timezone import now
from esi.errors import TokenExpiredError, TokenInvalidError
from esi.models import Token
from eveuniverse.models import EveEntity
from eveuniverse.tests.testdata.factories_2 import (
    EveEntityAllianceFactory,
    EveEntityCharacterFactory,
    EveEntityCorporationFactory,
    EveEntityFactionFactory,
)

from allianceauth.eveonline.models import EveCharacter
from app_utils.testdata_factories import EveCharacterFactory, UserMainFactory
from app_utils.testing import NoSocketsTestCase

from standingssync.core.esi_contacts import EsiContact, EsiContactsContainer
from standingssync.models import (
    EveContact,
    SyncedCharacter,
    SyncManager,
    _get_or_create_eve_entity_from_participant,
)
from standingssync.tests.factories import (
    EsiContactAllianceFactory,
    EsiContactCharacterFactory,
    EsiContactCorporationFactory,
    EsiContactFactionFactory,
    EsiContactLabelFactory,
    EveContactFactory,
    EveWarEmptyFactory,
    EveWarFactory,
    SyncedCharacterFactory,
    SyncManagerFactory,
    UserMainDefaultFactory,
    UserMainManagerFactory,
    make_esi_url,
)
from standingssync.tests.helpers import (
    EsiCharacterContactsStub,
    TestCaseWithClearCache,
    extract,
)

ESI_CONTACTS_PATH = "standingssync.core.esi_contacts"
ESI_API_PATH = "standingssync.core.esi_api"
MODELS_PATH = "standingssync.models"
WAR_TARGET_LABEL = "WAR TARGETS"


class TestSyncManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainManagerFactory()
        cls.alliance_id = cls.user.profile.main_character.alliance_id

    def test_should_return_token(self):
        # given
        obj = SyncManagerFactory()
        # when/then
        self.assertIsInstance(obj.fetch_token(), Token)

    def test_should_return_none_when_no_character_ownership(self):
        # given
        obj = SyncManagerFactory(character_ownership=None)
        # when/then
        self.assertIsNone(obj.fetch_token())

    def test_should_return_character(self):
        # given
        obj = SyncManagerFactory()
        # when/then
        self.assertEqual(obj.character, obj.character_ownership.character)  # type: ignore

    def test_should_raise_error_when_no_character(self):
        # given
        obj = SyncManagerFactory(character_ownership=None)
        # when
        with self.assertRaises(ValueError):
            _ = obj.character

    def test_should_report_sync_as_ok(self):
        # given
        my_dt = now()
        sync_manager = SyncManagerFactory(last_sync_at=my_dt - dt.timedelta(minutes=1))
        # when/then
        with patch(MODELS_PATH + ".STANDINGSSYNC_SYNC_TIMEOUT", 60):
            self.assertTrue(sync_manager.is_sync_fresh)

    def test_should_report_sync_as_not_ok(self):
        # given
        my_dt = now()
        sync_manager = SyncManagerFactory(last_sync_at=my_dt - dt.timedelta(minutes=61))
        # when/then
        with patch(MODELS_PATH + ".STANDINGSSYNC_SYNC_TIMEOUT", 60):
            self.assertFalse(sync_manager.is_sync_fresh)


def war_target_contact_ids(sync_manager: SyncManager) -> Set[int]:
    query = sync_manager.contacts.filter(is_war_target=True).values_list(
        "eve_entity_id", flat=True
    )
    return set(query)


@patch(MODELS_PATH + ".esi_api")
class TestSyncManager_RunSync(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainManagerFactory()
        cls.my_alliance_id = cls.user.profile.main_character.alliance_id

    def test_should_add_new_contacts_from_scratch_no_wt(self, mock_esi_api):
        # given
        contact_esi = EsiContactCharacterFactory()
        mock_esi_api.fetch_alliance_contacts.return_value = [contact_esi]
        EveEntityCharacterFactory(id=contact_esi.contact_id)
        sm = SyncManagerFactory(user=self.user)

        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False):
            sm.run_sync()

        # then
        self.assertEqual(sm.contacts.count(), 1)
        contact: EveContact = sm.contacts.first()
        self.assertEqual(contact.eve_entity.id, contact_esi.contact_id)
        self.assertEqual(contact.standing, contact_esi.standing)
        self.assertFalse(contact.is_war_target)

    def test_should_update_existing_contacts_no_wt(self, mock_esi_api):
        # given
        sm = SyncManagerFactory(user=self.user)
        contact = EveContactFactory(manager=sm, standing=-5)
        contact_esi: EsiContact = EsiContact.from_eve_entity(contact.eve_entity, 10.0)
        mock_esi_api.fetch_alliance_contacts.return_value = [contact_esi]

        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False):
            sm.run_sync()

        # then
        sm.refresh_from_db()
        self.assertEqual(sm.contacts.count(), 1)
        contact: EveContact = sm.contacts.first()
        self.assertEqual(contact.standing, 10.0)

    def test_should_not_update_contacts_when_unchanged(self, mock_esi_api):
        # given
        contact_esi = EsiContactCharacterFactory()
        mock_esi_api.fetch_alliance_contacts.return_value = [contact_esi]
        EveEntityCharacterFactory(id=contact_esi.contact_id)
        sm = SyncManagerFactory(user=self.user)

        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False):
            sm.run_sync()
            got = sm.run_sync()

        # then
        self.assertFalse(got)

    def test_should_update_contacts_when_unchanged_but_forced(self, mock_esi_api):
        # given
        contact_esi = EsiContactCharacterFactory()
        mock_esi_api.fetch_alliance_contacts.return_value = [contact_esi]
        EveEntityCharacterFactory(id=contact_esi.contact_id)
        sm = SyncManagerFactory(user=self.user)

        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False):
            sm.run_sync()
            got = sm.run_sync(force_update=True)

        # then
        self.assertTrue(got)

    def test_should_remove_obsolete_contacts_no_wt(self, mock_esi_api):
        # given
        sm = SyncManagerFactory(user=self.user)
        EveContactFactory(manager=sm)  # to be removed
        remaining_contact = EveContactFactory(manager=sm)
        contact = EsiContact.from_eve_contact(remaining_contact)
        mock_esi_api.fetch_alliance_contacts.return_value = [contact]

        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False):
            sm.run_sync()

        # then
        got = extract(sm.contacts, "eve_entity_id")
        want = {remaining_contact.contact_id}
        self.assertSetEqual(got, want)

    def test_should_add_new_contacts_from_scratch_with_wt(self, mock_esi_api):
        # given
        contact_esi = EsiContactCharacterFactory()
        mock_esi_api.fetch_alliance_contacts.return_value = [contact_esi]
        EveEntityCharacterFactory(id=contact_esi.contact_id)
        sm = SyncManagerFactory(user=self.user)
        aggressor = EveEntityAllianceFactory()
        defender = EveEntityAllianceFactory(id=self.my_alliance_id)
        EveWarFactory(aggressor=aggressor, defender=defender)
        EveWarFactory()  # should be ignored

        # when
        with (
            patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True),
            patch(
                MODELS_PATH + ".EsiContactsContainer.prune",
                lambda *args, **kwargs: 0,
            ),
        ):
            sm.run_sync()

        # then
        got = extract(sm.contacts, "eve_entity_id")
        want = {aggressor.id, contact_esi.contact_id}
        self.assertSetEqual(got, want)
        contact: EveContact = sm.contacts.get(eve_entity_id=aggressor.id)
        self.assertEqual(contact.standing, -10.0)
        self.assertTrue(contact.is_war_target)

    def test_should_add_war_target_contacts_as_aggressor(self, mock_esi_api):
        # given
        mock_esi_api.fetch_alliance_contacts.return_value = []
        sm = SyncManagerFactory(user=self.user)
        aggressor = EveEntityAllianceFactory(id=self.my_alliance_id)
        defender = EveEntityAllianceFactory()
        ally = EveEntityAllianceFactory()
        EveWarFactory(aggressor=aggressor, defender=defender, allies=[ally])
        EveWarFactory()  # should be ignored

        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True):
            sm.run_sync()

        # then
        got = extract(sm.contacts, "eve_entity_id")
        want = {defender.id, ally.id}
        self.assertSetEqual(got, want)

        contact_1: EveContact = sm.contacts.get(eve_entity_id=defender.id)
        self.assertEqual(contact_1.standing, -10.0)
        self.assertTrue(contact_1.is_war_target)

        contact_2: EveContact = sm.contacts.get(eve_entity_id=ally.id)
        self.assertEqual(contact_2.standing, -10.0)
        self.assertTrue(contact_2.is_war_target)

    def test_should_add_war_target_contact_as_defender(self, mock_esi_api):
        # given
        mock_esi_api.fetch_alliance_contacts.return_value = []
        sm = SyncManagerFactory(user=self.user)
        aggressor = EveEntityAllianceFactory()
        defender = EveEntityAllianceFactory(id=self.my_alliance_id)
        EveWarFactory(aggressor=aggressor, defender=defender)
        EveWarFactory()  # should be ignored

        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True):
            sm.run_sync()

        # then
        got = extract(sm.contacts, "eve_entity_id")
        want = {aggressor.id}
        self.assertSetEqual(got, want)
        contact: EveContact = sm.contacts.get(eve_entity_id=aggressor.id)
        self.assertEqual(contact.standing, -10.0)
        self.assertTrue(contact.is_war_target)

    def test_should_add_war_target_contact_as_ally(self, mock_esi_api):
        # given
        mock_esi_api.fetch_alliance_contacts.return_value = []
        sm = SyncManagerFactory(user=self.user)
        aggressor = EveEntityAllianceFactory()
        defender = EveEntityAllianceFactory()
        ally = EveEntityAllianceFactory(id=self.my_alliance_id)
        EveWarFactory(aggressor=aggressor, defender=defender, allies=[ally])
        EveWarFactory()  # should be ignored

        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True):
            sm.run_sync()

        # then
        got = extract(sm.contacts, "eve_entity_id")
        want = {aggressor.id}
        self.assertSetEqual(got, want)
        contact: EveContact = sm.contacts.get(eve_entity_id=aggressor.id)
        self.assertEqual(contact.standing, -10.0)
        self.assertTrue(contact.is_war_target)

    def test_remove_outdated_war_target_contacts(self, mock_esi_api):
        # given
        mock_esi_api.fetch_alliance_contacts.return_value = []
        sm = SyncManagerFactory(user=self.user)
        aggressor = EveEntityAllianceFactory()
        defender = EveEntityAllianceFactory(id=self.my_alliance_id)
        EveWarFactory(
            aggressor=aggressor,
            defender=defender,
            finished=now(),
        )
        EveContactFactory(manager=sm, eve_entity=aggressor, is_war_target=True)
        EveWarFactory()  # should be ignored

        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True):
            sm.run_sync()

        # then
        got = extract(sm.contacts, "eve_entity_id")
        want = set()
        self.assertSetEqual(got, want)
        self.assertFalse(sm.contacts.filter(eve_entity_id=aggressor.id).exists())

    def test_should_abort_when_no_char(self, mock_esi_api):
        # given
        sync_manager = SyncManagerFactory(character_ownership=None)
        # when/then
        with self.assertRaises(RuntimeError):
            sync_manager.run_sync()

    def test_should_abort_when_insufficient_permission(self, mock_esi_api):
        # given
        sync_manager = SyncManagerFactory(user=UserMainDefaultFactory())

        # when/then
        with self.assertRaises(RuntimeError):
            sync_manager.run_sync()

    def test_should_report_error_when_character_has_no_valid_token(self, mock_esi_api):
        # given
        user = UserMainManagerFactory()
        sync_manager = SyncManagerFactory(user=user)
        user.token_set.all().delete()

        # when/then
        with self.assertRaises(RuntimeError):
            sync_manager.run_sync()

    def test_should_call_prune_when_required(self, mock_esi_api):
        # given
        mock_esi_api.fetch_alliance_contacts.return_value = []

        class Case(NamedTuple):
            compress_contacts: bool
            unmask_war_targets: bool
            add_wt: bool
            want_called: bool
            want_unmask_war_targets: bool = None

        cases = [
            Case(
                compress_contacts=False,
                unmask_war_targets=False,
                add_wt=False,
                want_called=False,
            ),
            Case(
                compress_contacts=False,
                unmask_war_targets=False,
                add_wt=True,
                want_called=False,
            ),
            Case(
                compress_contacts=False,
                unmask_war_targets=True,
                add_wt=False,
                want_called=False,
            ),
            Case(
                compress_contacts=False,
                unmask_war_targets=True,
                add_wt=True,
                want_called=True,
                want_unmask_war_targets=True,
            ),
            Case(
                compress_contacts=True,
                unmask_war_targets=False,
                add_wt=False,
                want_called=True,
                want_unmask_war_targets=False,
            ),
            Case(
                compress_contacts=True,
                unmask_war_targets=False,
                add_wt=True,
                want_called=True,
                want_unmask_war_targets=False,
            ),
            Case(
                compress_contacts=True,
                unmask_war_targets=True,
                add_wt=False,
                want_called=True,
                want_unmask_war_targets=False,
            ),
            Case(
                compress_contacts=True,
                unmask_war_targets=True,
                add_wt=True,
                want_called=True,
                want_unmask_war_targets=True,
            ),
        ]

        # when
        for i, tc in enumerate(cases, start=1):
            # with self.subTest(i=i):
            with (
                patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", tc.add_wt),
                patch(MODELS_PATH + ".EsiContactsContainer.prune") as mock_prune,
            ):
                sm = SyncManagerFactory(
                    unmask_war_targets=tc.unmask_war_targets,
                    compress_contacts=tc.compress_contacts,
                )
                sm.run_sync()

                # then
                self.assertEqual(mock_prune.called, tc.want_called)
                if not tc.want_called:
                    continue

                _, kwargs = mock_prune.call_args
                self.assertEqual(kwargs["compress_contacts"], tc.compress_contacts)
                self.assertEqual(
                    kwargs["unmask_war_targets"], tc.want_unmask_war_targets
                )


@patch(MODELS_PATH + ".esi_api")
class TestSyncManager_RunSync_Filters(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainManagerFactory()

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False)
    def test_should_apply_contact_filters_no_wt(self, mock_esi_api):
        # given
        alliance_bad = EsiContactAllianceFactory(contact_id=3011, standing=-5)
        alliance_good = EsiContactAllianceFactory(contact_id=3001, standing=5)
        alliance_neutral = EsiContactAllianceFactory(contact_id=3002, standing=0)
        character_bad = EsiContactCharacterFactory(contact_id=1011, standing=-5)
        character_good = EsiContactCharacterFactory(contact_id=1001, standing=5)
        character_neutral = EsiContactCharacterFactory(contact_id=1002, standing=0)
        corporation_bad = EsiContactCorporationFactory(contact_id=2011, standing=-5)
        corporation_good = EsiContactCorporationFactory(contact_id=2001, standing=5)
        corporation_neutral = EsiContactCorporationFactory(contact_id=2002, standing=0)
        faction_bad = EsiContactFactionFactory(contact_id=5011, standing=-5)
        faction_good = EsiContactFactionFactory(contact_id=5001, standing=5)
        faction_neutral = EsiContactFactionFactory(contact_id=5002, standing=0)
        contacts = [
            alliance_bad,
            alliance_good,
            alliance_neutral,
            character_bad,
            character_good,
            character_neutral,
            corporation_bad,
            corporation_good,
            corporation_neutral,
            faction_bad,
            faction_good,
            faction_neutral,
        ]

        mock_esi_api.fetch_alliance_contacts.return_value = contacts
        for c in contacts:
            match c.contact_type:
                case EsiContact.Category.ALLIANCE:
                    EveEntityAllianceFactory(id=c.contact_id)
                case EsiContact.Category.CHARACTER:
                    EveEntityCharacterFactory(id=c.contact_id)
                case EsiContact.Category.CORPORATION:
                    EveEntityCorporationFactory(id=c.contact_id)
                case EsiContact.Category.FACTION:
                    EveEntityFactionFactory(id=c.contact_id)

        class Case(NamedTuple):
            exclude_character_contacts: bool
            exclude_positive_standings: bool
            exclude_negative_standings: bool
            want: List[EsiContact]

        cases = [
            Case(
                exclude_character_contacts=False,
                exclude_positive_standings=False,
                exclude_negative_standings=False,
                want=[
                    alliance_bad,
                    alliance_good,
                    alliance_neutral,
                    character_bad,
                    character_good,
                    character_neutral,
                    corporation_bad,
                    corporation_good,
                    corporation_neutral,
                    faction_bad,
                    faction_good,
                    faction_neutral,
                ],
            ),
            Case(
                exclude_character_contacts=True,
                exclude_positive_standings=False,
                exclude_negative_standings=False,
                want=[
                    alliance_bad,
                    alliance_good,
                    alliance_neutral,
                    corporation_bad,
                    corporation_good,
                    corporation_neutral,
                    faction_bad,
                    faction_good,
                    faction_neutral,
                ],
            ),
            Case(
                exclude_character_contacts=False,
                exclude_positive_standings=True,
                exclude_negative_standings=False,
                want=[
                    alliance_bad,
                    alliance_neutral,
                    character_bad,
                    character_neutral,
                    corporation_bad,
                    corporation_neutral,
                    faction_bad,
                    faction_neutral,
                ],
            ),
            Case(
                exclude_character_contacts=False,
                exclude_positive_standings=False,
                exclude_negative_standings=True,
                want=[
                    alliance_good,
                    alliance_neutral,
                    character_good,
                    character_neutral,
                    corporation_good,
                    corporation_neutral,
                    faction_good,
                    faction_neutral,
                ],
            ),
            Case(
                exclude_character_contacts=False,
                exclude_positive_standings=True,
                exclude_negative_standings=True,
                want=[
                    alliance_neutral,
                    character_neutral,
                    corporation_neutral,
                    faction_neutral,
                ],
            ),
            Case(
                exclude_character_contacts=True,
                exclude_positive_standings=True,
                exclude_negative_standings=False,
                want=[
                    alliance_bad,
                    alliance_neutral,
                    corporation_bad,
                    corporation_neutral,
                    faction_bad,
                    faction_neutral,
                ],
            ),
            Case(
                exclude_character_contacts=True,
                exclude_positive_standings=False,
                exclude_negative_standings=True,
                want=[
                    alliance_good,
                    alliance_neutral,
                    corporation_good,
                    corporation_neutral,
                    faction_good,
                    faction_neutral,
                ],
            ),
            Case(
                exclude_character_contacts=True,
                exclude_positive_standings=True,
                exclude_negative_standings=True,
                want=[
                    alliance_neutral,
                    corporation_neutral,
                    faction_neutral,
                ],
            ),
        ]
        for i, tc in enumerate(cases, start=1):
            with self.subTest(num=i):
                sm = SyncManagerFactory(
                    user=self.user,
                    exclude_character_contacts=tc.exclude_character_contacts,
                    exclude_positive_standings=tc.exclude_positive_standings,
                    exclude_negative_standings=tc.exclude_negative_standings,
                )

                # when
                sm.run_sync()

                # then
                got = set(sm.contacts.values_list("eve_entity__id", flat=True))
                want = {x.contact_id for x in tc.want}
                self.assertSetEqual(got, want)
                sm.delete()


class TestSyncManager_AddWarTargets(NoSocketsTestCase):
    def test_should_add_war_targets(self):
        # given
        sync_manager = SyncManagerFactory()
        alliance_entity = EveEntityAllianceFactory(
            id=sync_manager.alliance.alliance_id,
            name=sync_manager.alliance.alliance_name,
        )
        alliance_contacts = EsiContactsContainer()
        war = EveWarFactory(defender=alliance_entity)
        aggressor = EsiContact.from_eve_entity(
            war.aggressor, standing=-10, is_war_target=True
        )
        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True):
            result = sync_manager._add_war_targets(alliance_contacts)
        # then
        self.assertSetEqual(alliance_contacts.contacts(), {aggressor})
        self.assertSetEqual(result, {aggressor.contact_id})

    def test_should_not_add_war_targets_when_disabled(self):
        # given
        sync_manager = SyncManagerFactory()
        alliance_entity = EveEntityAllianceFactory(
            id=sync_manager.alliance.alliance_id,
            name=sync_manager.alliance.alliance_name,
        )
        alliance_contacts = EsiContactsContainer()
        EveWarFactory(defender=alliance_entity)
        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False):
            result = sync_manager._add_war_targets(alliance_contacts)
        # then
        self.assertSetEqual(alliance_contacts.contacts(), set())
        self.assertSetEqual(result, set())

    def test_should_ignore_unknown_war_targets(self):
        # given
        sync_manager = SyncManagerFactory()
        alliance_entity = EveEntityAllianceFactory(
            id=sync_manager.alliance.alliance_id,
            name=sync_manager.alliance.alliance_name,
        )
        alliance_contacts = EsiContactsContainer()
        ally = EveEntity.objects.create(id=1234567)
        war = EveWarFactory(aggressor=alliance_entity, allies=[ally])
        defender = EsiContact.from_eve_entity(
            war.defender, standing=-10, is_war_target=True
        )
        # when
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True):
            result = sync_manager._add_war_targets(alliance_contacts)
        # then
        self.assertSetEqual(alliance_contacts.contacts(), {defender})
        self.assertSetEqual(result, {defender.contact_id})


class TestSyncManager_EffectiveStandingWithCharacter(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.sync_manager = SyncManagerFactory()
        contacts = [
            (EveEntityCharacterFactory(id=1001), -10),
            (EveEntityCorporationFactory(id=2001), 10),
            (EveEntityAllianceFactory(id=3001), 5),
        ]
        for contact, standing in contacts:
            EveContactFactory(
                manager=cls.sync_manager, eve_entity=contact, standing=standing
            )

    def test_char_with_character_standing(self):
        c1 = EveCharacter(
            character_id=1001,
            character_name="Char 1",
            corporation_id=201,
            corporation_name="Corporation 1",
            corporation_ticker="C1",
        )
        self.assertEqual(self.sync_manager.effective_standing_with_character(c1), -10)

    def test_char_with_corporation_standing(self):
        c2 = EveCharacter(
            character_id=1002,
            character_name="Char 2",
            corporation_id=2001,
            corporation_name="Corporation 1",
            corporation_ticker="C1",
        )
        self.assertEqual(self.sync_manager.effective_standing_with_character(c2), 10)

    def test_char_with_alliance_standing(self):
        c3 = EveCharacter(
            character_id=1003,
            character_name="Char 3",
            corporation_id=2003,
            corporation_name="Corporation 3",
            corporation_ticker="C2",
            alliance_id=3001,
            alliance_name="Alliance 1",
            alliance_ticker="A1",
        )
        self.assertEqual(self.sync_manager.effective_standing_with_character(c3), 5)

    def test_char_without_standing_and_has_alliance(self):
        c4 = EveCharacter(
            character_id=1003,
            character_name="Char 3",
            corporation_id=2003,
            corporation_name="Corporation 3",
            corporation_ticker="C2",
            alliance_id=3002,
            alliance_name="Alliance 2",
            alliance_ticker="A2",
        )
        self.assertEqual(self.sync_manager.effective_standing_with_character(c4), 0.0)

    def test_char_without_standing_and_without_alliance_1(self):
        c4 = EveCharacter(
            character_id=1003,
            character_name="Char 3",
            corporation_id=2003,
            corporation_name="Corporation 3",
            corporation_ticker="C2",
            alliance_id=None,
            alliance_name=None,
            alliance_ticker=None,
        )
        self.assertEqual(self.sync_manager.effective_standing_with_character(c4), 0.0)

    def test_char_without_standing_and_without_alliance_2(self):
        c4 = EveCharacter(
            character_id=1003,
            character_name="Char 3",
            corporation_id=2003,
            corporation_name="Corporation 3",
            corporation_ticker="C2",
        )
        self.assertEqual(self.sync_manager.effective_standing_with_character(c4), 0.0)


@patch(MODELS_PATH + ".notify")
class TestSyncCharacter_FetchToken(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.sync_manager = SyncManagerFactory()

    def test_should_return_token(self, mock_notify):
        # given
        sc = SyncedCharacterFactory(manager=self.sync_manager)

        # when
        result = sc.fetch_token()

        # then
        self.assertIsInstance(result, Token)
        self.assertFalse(mock_notify.called)

    def test_should_return_none_and_delete_when_token_invalid(self, mock_notify):
        # given
        obj = SyncedCharacterFactory(manager=self.sync_manager)
        with patch(MODELS_PATH + ".SyncedCharacter._valid_token") as m:
            m.side_effect = TokenInvalidError
            # when
            result = obj.fetch_token()
        # then
        self.assertIsNone(result)
        self.assertFalse(SyncedCharacter.objects.filter(pk=obj.pk).exists())
        self.assertTrue(mock_notify.called)

    def test_should_return_none_and_delete_when_token_has_issues(self, mock_notify):
        params = [TokenInvalidError, TokenExpiredError]
        for exception in params:
            with self.subTest(exception=exception):
                # given
                obj = SyncedCharacterFactory(manager=self.sync_manager)
                # when
                with patch(MODELS_PATH + ".SyncedCharacter._valid_token") as m:
                    m.side_effect = exception
                    result = obj.fetch_token()
                # then
                self.assertIsNone(result)
                self.assertFalse(SyncedCharacter.objects.filter(pk=obj.pk).exists())
                self.assertTrue(mock_notify.called)

    def test_should_return_none_and_delete_when_token_not_found(self, mock_notify):
        # given
        obj = SyncedCharacterFactory(manager=self.sync_manager)
        # when
        with patch(MODELS_PATH + ".SyncedCharacter._valid_token") as m:
            m.return_value = None
            result = obj.fetch_token()
        # then
        self.assertIsNone(result)
        self.assertFalse(SyncedCharacter.objects.filter(pk=obj.pk).exists())
        self.assertTrue(mock_notify.called)


@patch(ESI_CONTACTS_PATH + ".STANDINGSSYNC_WAR_TARGETS_LABEL_NAME", WAR_TARGET_LABEL)
@patch(MODELS_PATH + ".notify")
@patch(MODELS_PATH + ".esi_api")
class TestSyncCharacter_RunSync(NoSocketsTestCase):
    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    def test_should_replace_contacts_no_wt(self, mock_esi_api, mock_notify):
        # given
        synced_character = SyncedCharacterFactory()
        sync_manager = synced_character.manager
        EveContactFactory(
            manager=sync_manager,
            eve_entity=EveEntityCharacterFactory(id=synced_character.character_id),
            standing=10,
        )
        alliance_contact_1 = EveContactFactory(manager=sync_manager)
        alliance_contact_2 = EveContactFactory(manager=sync_manager, standing=10)
        character_contact_1 = EsiContact.from_eve_entity(
            EveEntityCharacterFactory(), -5
        )
        EveContactFactory(
            manager=sync_manager, standing=-10, is_war_target=True
        )  # alliance_wt_contact
        character_contact_2 = EsiContact.from_eve_contact(alliance_contact_2).clone(
            standing=-5
        )
        esi_character_contacts = EsiCharacterContactsStub.create(
            synced_character.character_id,
            mock_esi_api,
            contacts=[character_contact_1, character_contact_2],
        )
        # when
        result = synced_character.run_sync()
        # then
        self.assertTrue(result)
        synced_character.refresh_from_db()
        self.assertIsNotNone(synced_character.last_sync_at)
        expected = {
            EsiContact.from_eve_contact(alliance_contact_1),
            EsiContact.from_eve_contact(alliance_contact_2),
        }
        self.assertSetEqual(esi_character_contacts.contacts(), expected)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0)
    def test_should_replace_contacts_no_wt_no_standing(self, mock_esi_api, mock_notify):
        # given
        synced_character = SyncedCharacterFactory()
        sync_manager = synced_character.manager
        EveContactFactory(
            manager=sync_manager,
            eve_entity=EveEntityCharacterFactory(id=synced_character.character_id),
            standing=0,
        )
        alliance_contact_1 = EveContactFactory(manager=sync_manager)
        alliance_contact_2 = EveContactFactory(manager=sync_manager, standing=10)
        character_contact_1 = EsiContact.from_eve_entity(
            EveEntityCharacterFactory(), -5
        )
        EveContactFactory(
            manager=sync_manager, standing=-10, is_war_target=True
        )  # alliance_wt_contact
        character_contact_2 = EsiContact.from_eve_contact(alliance_contact_2).clone(
            standing=-5
        )
        esi_character_contacts = EsiCharacterContactsStub.create(
            synced_character.character_id,
            mock_esi_api,
            contacts=[character_contact_1, character_contact_2],
        )
        # when
        result = synced_character.run_sync()
        # then
        self.assertTrue(result)
        synced_character.refresh_from_db()
        self.assertIsNotNone(synced_character.last_sync_at)
        expected = {
            EsiContact.from_eve_contact(alliance_contact_1),
            EsiContact.from_eve_contact(alliance_contact_2),
        }
        self.assertSetEqual(esi_character_contacts.contacts(), expected)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    def test_should_replace_contacts_incl_wt(self, mock_esi_api, mock_notify):
        # given
        synced_character = SyncedCharacterFactory()
        sync_manager = synced_character.manager
        EveContactFactory(
            manager=sync_manager,
            eve_entity=EveEntityCharacterFactory(id=synced_character.character_id),
            standing=10,
        )
        alliance_contact_1 = EveContactFactory(manager=sync_manager)
        alliance_contact_2 = EveContactFactory(manager=sync_manager, standing=10)
        alliance_wt_contact = EveContactFactory(
            manager=sync_manager, standing=-10, is_war_target=True
        )
        character_contact_1 = EsiContact.from_eve_entity(
            EveEntityCharacterFactory(), -5
        )
        character_contact_2 = EsiContact.from_eve_contact(alliance_contact_2).clone(
            standing=-5
        )
        wt_label = EsiContactLabelFactory(name=WAR_TARGET_LABEL)
        esi_character_contacts = EsiCharacterContactsStub.create(
            synced_character.character_id,
            mock_esi_api,
            contacts=[character_contact_1, character_contact_2],
            labels=[wt_label],
        )
        # when
        result = synced_character.run_sync()
        # then
        self.assertTrue(result)
        synced_character.refresh_from_db()
        self.assertIsNotNone(synced_character.last_sync_at)
        expected = {
            EsiContact.from_eve_contact(alliance_contact_1),
            EsiContact.from_eve_contact(alliance_contact_2),
            EsiContact.from_eve_contact(alliance_wt_contact, label_ids=[wt_label.id]),
        }
        self.assertSetEqual(esi_character_contacts.contacts(), expected)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    def test_should_not_update_anything(self, mock_esi_api, mock_notify):
        # given
        synced_character = SyncedCharacterFactory()
        sync_manager = synced_character.manager
        EveContactFactory(
            manager=sync_manager,
            eve_entity=EveEntityCharacterFactory(id=synced_character.character_id),
            standing=10,
        )
        EveContactFactory(manager=sync_manager)  # alliance_contact_1
        alliance_contact_2 = EveContactFactory(manager=sync_manager, standing=10)
        character_contact_1 = EsiContact.from_eve_entity(
            EveEntityCharacterFactory(), -5
        )
        EveContactFactory(
            manager=sync_manager, standing=-10, is_war_target=True
        )  # alliance_wt_contact
        character_contact_2 = EsiContact.from_eve_contact(alliance_contact_2).clone(
            standing=-5
        )
        esi_character_contacts = EsiCharacterContactsStub.create(
            synced_character.character_id,
            mock_esi_api,
            contacts=[character_contact_1, character_contact_2],
        )
        # when
        result = synced_character.run_sync()
        # then
        self.assertTrue(result)
        synced_character.refresh_from_db()
        self.assertIsNotNone(synced_character.last_sync_at)
        expected = {character_contact_1, character_contact_2}
        self.assertSetEqual(esi_character_contacts.contacts(), expected)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    def test_should_sync_war_targets_but_not_alliance_contacts(
        self, mock_esi_api, mock_notify
    ):
        # given
        synced_character = SyncedCharacterFactory()
        sync_manager = synced_character.manager
        wt_label = EsiContactLabelFactory(name=WAR_TARGET_LABEL)
        EveContactFactory(manager=sync_manager)  # should not sync this alliance contact
        EveContactFactory(
            manager=sync_manager,
            eve_entity=EveEntityCharacterFactory(id=synced_character.character_id),
            standing=10,
        )  # sync char must have standing with alliance
        alliance_wt_contact_1 = EveContactFactory(
            manager=sync_manager, standing=-10, is_war_target=True
        )  # new war target
        alliance_wt_contact_2 = EveContactFactory(
            manager=sync_manager, standing=-10, is_war_target=True
        )  # new war target
        character_contact_1 = EsiContact.from_eve_entity(
            EveEntityCharacterFactory(), -5
        )  # character contact must be kept in place
        character_old_wt_contact = EsiContact.from_eve_entity(
            EveEntityCharacterFactory(), -10, [wt_label.id]
        )  # should remove this old WT contact
        character_contact_2 = EsiContact.from_eve_contact(alliance_wt_contact_2).clone(
            standing=10
        )  # should replace this existing character contact with a WT
        esi_character_contacts = EsiCharacterContactsStub.create(
            synced_character.character_id,
            mock_esi_api,
            contacts=[
                character_contact_1,
                character_old_wt_contact,
                character_contact_2,
            ],
            labels=[wt_label],
        )
        # when
        result = synced_character.run_sync()
        # then
        self.assertTrue(result)
        synced_character.refresh_from_db()
        self.assertIsNotNone(synced_character.last_sync_at)
        expected = {
            character_contact_1,
            EsiContact.from_eve_contact(alliance_wt_contact_1, label_ids=[wt_label.id]),
            EsiContact.from_eve_contact(alliance_wt_contact_2, label_ids=[wt_label.id]),
        }
        self.assertSetEqual(esi_character_contacts.contacts(), expected)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    def test_should_add_wt_label_info(self, mock_esi_api, mock_notify):
        # given
        synced_character = SyncedCharacterFactory()
        sync_manager = synced_character.manager
        EveContactFactory(
            manager=sync_manager,
            eve_entity=EveEntityCharacterFactory(id=synced_character.character_id),
            standing=10,
        )
        EveContactFactory(manager=sync_manager)  # alliance_contact
        wt_label = EsiContactLabelFactory(name=WAR_TARGET_LABEL)
        EsiCharacterContactsStub.create(
            synced_character.character_id, mock_esi_api, labels=[wt_label]
        )
        # when
        synced_character.run_sync()
        # then
        synced_character.refresh_from_db()
        self.assertTrue(synced_character.has_war_targets_label)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    def test_should_remove_wt_label_info(self, mock_esi_api, mock_notify):
        # given
        synced_character = SyncedCharacterFactory(has_war_targets_label=True)
        sync_manager = synced_character.manager
        EveContactFactory(
            manager=sync_manager,
            eve_entity=EveEntityCharacterFactory(id=synced_character.character_id),
            standing=10,
        )
        EveContactFactory(manager=sync_manager)  # alliance_contact
        other_label = EsiContactLabelFactory()
        EsiCharacterContactsStub.create(
            synced_character.character_id, mock_esi_api, labels=[other_label]
        )
        # when
        synced_character.run_sync()
        # then
        synced_character.refresh_from_db()
        self.assertFalse(synced_character.has_war_targets_label)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    def test_should_not_sync_when_no_contacts(self, mock_esi_api, mock_notify):
        # given
        synced_character = SyncedCharacterFactory()
        sync_manager = synced_character.manager
        EveContactFactory(
            manager=sync_manager,
            eve_entity=EveEntityCharacterFactory(id=synced_character.character_id),
            standing=10,
        )
        character_contact_1 = EsiContact.from_eve_entity(
            EveEntityCharacterFactory(), -5
        )
        EsiCharacterContactsStub.create(
            synced_character.character_id,
            mock_esi_api,
            contacts=[character_contact_1],
        )
        # when
        result = synced_character.run_sync()
        # then
        self.assertIsNone(result)

    def test_should_delete_contacts(self, mock_esi_api, mock_notify):
        # given
        synced_character = SyncedCharacterFactory()
        character_contact_1 = EsiContactCharacterFactory()
        esi_character_contacts = EsiCharacterContactsStub.create(
            synced_character.character_id,
            mock_esi_api,
            contacts=[character_contact_1],
        )
        # when
        synced_character.delete_all_contacts()
        # then
        self.assertSetEqual(esi_character_contacts.contacts(), set())

    def test_should_do_nothing_when_no_token(self, mock_esi_api, mock_notify):
        # given
        obj = SyncedCharacterFactory()
        obj.character_ownership.user.token_set.all().delete()
        with patch(MODELS_PATH + ".esi_api.delete_character_contacts") as esi_api:
            # when
            obj.delete_all_contacts()
            # then
            self.assertFalse(esi_api.called)

    def test_should_delete_when_insufficient_permission(
        self, mock_esi_api, mock_notify
    ):
        # given
        sm = SyncManagerFactory()
        character = EveCharacterFactory(corporation__alliance=sm.alliance)
        user = UserMainFactory(main_character__character=character)
        sc = SyncedCharacterFactory(manager=sm, user=user)
        # when
        result = sc.run_sync()

        # then
        self.assertFalse(result)
        self.assertFalse(SyncedCharacter.objects.filter(pk=sc.pk).exists())
        self.assertTrue(mock_notify.called)

    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.1)
    def test_should_delete_when_character_has_no_standing(
        self, mock_esi_api, mock_notify
    ):
        # given
        sm = SyncManagerFactory()
        sc = SyncedCharacterFactory(manager=sm)
        EveContactFactory(
            manager=sm,
            eve_entity=EveEntityCharacterFactory(id=sc.character_id),
            standing=-10,
        )

        # when
        result = sc.run_sync()

        # then
        self.assertFalse(result)
        self.assertFalse(SyncedCharacter.objects.filter(pk=sc.pk).exists())
        self.assertTrue(mock_notify.called)


class TestSyncCharacter_IsSyncFresh(NoSocketsTestCase):
    def test_should_report_sync_as_ok(self):
        # given
        my_dt = now()
        obj = SyncedCharacterFactory(last_sync_at=my_dt - dt.timedelta(minutes=1))
        # when/then
        with patch(MODELS_PATH + ".STANDINGSSYNC_SYNC_TIMEOUT", 60):
            self.assertTrue(obj.is_sync_fresh)

    def test_should_report_sync_as_not_ok(self):
        # given
        my_dt = now()
        obj = SyncedCharacterFactory(last_sync_at=my_dt - dt.timedelta(minutes=61))
        # when/then
        with patch(MODELS_PATH + ".STANDINGSSYNC_SYNC_TIMEOUT", 60):
            self.assertFalse(obj.is_sync_fresh)


class TestSyncCharacter_UpdateWtLabelInfo(NoSocketsTestCase):
    def test_should_update_wt_label_info(self):
        # given
        synced_character = SyncedCharacterFactory()
        wt_label = EsiContactLabelFactory(name=WAR_TARGET_LABEL)
        character_contacts = EsiContactsContainer.from_esi_contacts(labels=[wt_label])
        # when
        synced_character._update_wt_label_info(character_contacts)
        # then
        synced_character.refresh_from_db()
        self.assertTrue(synced_character.has_war_targets_label)

    def test_should_not_update_wt_label_info(self):
        # given
        synced_character = SyncedCharacterFactory(has_war_targets_label=True)
        wt_label = EsiContactLabelFactory(name=WAR_TARGET_LABEL)
        character_contacts = EsiContactsContainer.from_esi_contacts(labels=[wt_label])
        # when
        synced_character._update_wt_label_info(character_contacts)
        # then
        synced_character.refresh_from_db()
        self.assertTrue(synced_character.has_war_targets_label)


class TestEveContact(NoSocketsTestCase):
    def test_str(self):
        # given
        contact = EveContactFactory(eve_entity__name="Alpha")
        # when/then
        self.assertEqual(str(contact), "Alpha")


class TestEveWar(NoSocketsTestCase):
    def test_str(self):
        # given
        aggressor = EveEntityAllianceFactory(name="Alpha")
        defender = EveEntityAllianceFactory(name="Bravo")
        war = EveWarFactory(aggressor=aggressor, defender=defender)
        # when/then
        self.assertEqual(str(war), "Alpha vs. Bravo")


class TestEveWar_UpdateFromESI(TestCaseWithClearCache):
    @pook.on
    def test_should_update_full_war_from_esi(self):
        # given
        war = EveWarEmptyFactory()
        declared = now() - dt.timedelta(days=5)
        started = now() - dt.timedelta(days=4)
        finished = now() + dt.timedelta(days=1)
        retracted = now()
        aggressor = EveEntityAllianceFactory()
        ally_1 = EveEntityAllianceFactory()
        ally_2 = EveEntityCorporationFactory()
        defender = EveEntityAllianceFactory()
        pook.get(
            make_esi_url(f"wars/{war.id}"),
            reply=HTTPStatus.OK,
            response_json={
                "aggressor": {
                    "alliance_id": aggressor.id,
                    "isk_destroyed": 0,
                    "ships_killed": 0,
                },
                "allies": [{"alliance_id": ally_1.id}, {"corporation_id": ally_2.id}],
                "declared": declared.isoformat(),
                "defender": {
                    "alliance_id": defender.id,
                    "isk_destroyed": 0,
                    "ships_killed": 0,
                },
                "finished": finished.isoformat(),
                "id": war.id,
                "mutual": False,
                "open_for_allies": True,
                "retracted": retracted.isoformat(),
                "started": started.isoformat(),
            },
        )

        # when
        war.update_from_esi()

        # then
        war.refresh_from_db()
        self.assertEqual(war.aggressor, aggressor)
        self.assertCountEqual(war.allies.all(), [ally_1, ally_2])
        self.assertEqual(war.declared, declared)
        self.assertEqual(war.defender, defender)
        self.assertEqual(war.finished, finished)
        self.assertFalse(war.is_mutual)
        self.assertTrue(war.is_open_for_allies)
        self.assertEqual(war.retracted, retracted)
        self.assertEqual(war.started, started)

    @pook.on
    def test_should_update_minimal_war_from_esi(self):
        # given
        war = EveWarEmptyFactory()
        declared = now() - dt.timedelta(days=5)
        started = now() - dt.timedelta(days=4)
        aggressor = EveEntityAllianceFactory()
        defender = EveEntityAllianceFactory()
        pook.get(
            make_esi_url(f"wars/{war.id}"),
            reply=HTTPStatus.OK,
            response_json={
                "aggressor": {
                    "alliance_id": aggressor.id,
                    "isk_destroyed": 0,
                    "ships_killed": 0,
                },
                "declared": declared.isoformat(),
                "defender": {
                    "alliance_id": defender.id,
                    "isk_destroyed": 0,
                    "ships_killed": 0,
                },
                "id": war.id,
                "mutual": False,
                "open_for_allies": True,
                "started": started.isoformat(),
            },
        )

        # when
        war.update_from_esi()

        # then
        war.refresh_from_db()
        self.assertEqual(war.aggressor, aggressor)
        self.assertEqual(war.allies.count(), 0)
        self.assertEqual(war.declared, declared)
        self.assertEqual(war.defender, defender)
        self.assertIsNone(war.finished)
        self.assertFalse(war.is_mutual)
        self.assertTrue(war.is_open_for_allies)
        self.assertIsNone(war.retracted)
        self.assertEqual(war.started, started)

    @pook.on
    def test_should_update_existing_war_from_esi(self):
        # given
        war = EveWarFactory()
        declared = now() - dt.timedelta(days=5)
        started = now() - dt.timedelta(days=4)
        finished = now() + dt.timedelta(days=1)
        retracted = now()
        aggressor = EveEntityAllianceFactory()
        ally_1 = EveEntityAllianceFactory()
        ally_2 = EveEntityCorporationFactory()
        defender = EveEntityAllianceFactory()
        pook.get(
            make_esi_url(f"wars/{war.id}"),
            reply=HTTPStatus.OK,
            response_json={
                "aggressor": {
                    "alliance_id": aggressor.id,
                    "isk_destroyed": 0,
                    "ships_killed": 0,
                },
                "allies": [{"alliance_id": ally_1.id}, {"corporation_id": ally_2.id}],
                "declared": declared.isoformat(),
                "defender": {
                    "alliance_id": defender.id,
                    "isk_destroyed": 0,
                    "ships_killed": 0,
                },
                "finished": finished.isoformat(),
                "id": war.id,
                "mutual": False,
                "open_for_allies": True,
                "retracted": retracted.isoformat(),
                "started": started.isoformat(),
            },
        )

        # when
        war.update_from_esi()

        # then
        war.refresh_from_db()
        self.assertEqual(war.aggressor, aggressor)
        self.assertCountEqual(war.allies.all(), [ally_1, ally_2])
        self.assertEqual(war.declared, declared)
        self.assertEqual(war.defender, defender)
        self.assertEqual(war.finished, finished)
        self.assertFalse(war.is_mutual)
        self.assertTrue(war.is_open_for_allies)
        self.assertEqual(war.retracted, retracted)
        self.assertEqual(war.started, started)


class TestEveWarManager__GetOrCreateEveEntityFromParticipant(NoSocketsTestCase):
    def test_should_create_from_alliance_id(self):
        # given
        alliance = EveEntityAllianceFactory()
        data = {"alliance_id": alliance.id}
        # when
        result = _get_or_create_eve_entity_from_participant(data)
        # then
        self.assertEqual(result, alliance)

    def test_should_create_from_corporation_id(self):
        # given
        corporation = EveEntityCorporationFactory()
        data = {"corporation_id": corporation.id}
        # when
        result = _get_or_create_eve_entity_from_participant(data)
        # then
        self.assertEqual(result, corporation)

    def test_should_raise_error_when_no_id_found(self):
        # given
        data = {}
        # when/then
        with self.assertRaises(ValueError):
            _get_or_create_eve_entity_from_participant(data)



# Generated by Django 5.2.14 on 2026-06-03 16:02

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0012_alter_evebloodline_eve_ship_type"),
        ("standingssync", "0002_improve_sync_logic"),
    ]

    operations = [
        migrations.AddField(
            model_name="evewar",
            name="last_modified",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="evewar",
            name="aggressor",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evewar",
            name="declared",
            field=models.DateTimeField(default=None, null=True),
        ),
        migrations.AlterField(
            model_name="evewar",
            name="defender",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evewar",
            name="is_mutual",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AlterField(
            model_name="evewar",
            name="is_open_for_allies",
            field=models.BooleanField(default=None, null=True),
        ),
    ]



"""Routes for standingssync."""

from django.urls import path

from standingssync import views

app_name = "standingssync"

urlpatterns = [
    path("", views.index, name="index"),
    path("characters", views.characters, name="characters"),
    path("add-character", views.add_character, name="add_character"),
    path(
        "remove-character/<int:alt_pk>", views.remove_character, name="remove_character"
    ),
    path(
        "add-alliance-manager", views.add_alliance_manager, name="add_alliance_manager"
    ),
    path("wars", views.wars, name="wars"),
    path("admin-update-wars", views.admin_update_wars, name="admin_update_wars"),
]



from django.utils.translation import gettext_lazy as _

from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

from standingssync import __title__, urls


class standingssyncMenuItem(MenuItemHook):
    """This class ensures only authorized users will see the menu entry"""

    def __init__(self):
        # setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            _(__title__),
            "fas fa-sync-alt",
            "standingssync:characters",
            navactive=["standingssync:characters"],
        )

    def render(self, request):
        if request.user.has_perm("standingssync.add_syncedcharacter"):
            return MenuItemHook.render(self, request)
        return ""


@hooks.register("menu_item_hook")
def register_menu():
    return standingssyncMenuItem()


@hooks.register("url_hook")
def register_urls():
    return UrlHook(urls, "standingssync", r"^standingssync/")



# Generated by Django 5.2.15 on 2026-06-23 12:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("standingssync", "0004_add_contact_filters_to_manager"),
    ]

    operations = [
        migrations.AddField(
            model_name="syncmanager",
            name="unmask_war_targets",
            field=models.BooleanField(
                default=False,
                help_text="Remove contacts which would mask the effective standing of war target contacts",
            ),
        ),
        migrations.AlterField(
            model_name="syncmanager",
            name="compress_contacts",
            field=models.BooleanField(
                default=False,
                help_text="Remove contacts which are unnecessary for calculating their effective standing.",
            ),
        ),
        migrations.AlterField(
            model_name="syncmanager",
            name="exclude_character_contacts",
            field=models.BooleanField(
                default=False, help_text="Exclude all character contacts"
            ),
        ),
        migrations.AlterField(
            model_name="syncmanager",
            name="exclude_negative_standings",
            field=models.BooleanField(
                default=False, help_text="Exclude all contacts with negative standing"
            ),
        ),
        migrations.AlterField(
            model_name="syncmanager",
            name="exclude_positive_standings",
            field=models.BooleanField(
                default=False, help_text="Exclude all contacts with positive standing"
            ),
        ),
    ]



"""Core logic for handling ESI contacts."""

import hashlib
import json
from copy import deepcopy
from dataclasses import dataclass, field, fields
from enum import Enum
from typing import Any, Dict, FrozenSet, Iterable, List, Optional, Set, Tuple

from eveuniverse.models import EveEntity

from app_utils.helpers import chunks

from standingssync.app_settings import STANDINGSSYNC_WAR_TARGETS_LABEL_NAME
from standingssync.providers import esi


@dataclass(frozen=True)
class EsiContactLabel:
    """An ESI contact label. Immutable."""

    id: int
    name: str

    def __post_init__(self):
        object.__setattr__(self, "id", int(self.id))
        object.__setattr__(self, "name", str(self.name))

    def to_dict(self) -> dict:
        """Return as dict."""
        return {self.id: self.name}

    def to_esi_dict(self) -> dict:
        """Return as dict in ESI format."""
        return {"label_id": self.id, "label_name": self.name}

    @classmethod
    def from_esi_dict(cls, esi_dict: dict) -> "EsiContactLabel":
        """Create new obj from ESI dict."""
        return cls(id=esi_dict["label_id"], name=esi_dict["label_name"])


@dataclass(frozen=True)
class EsiContact:
    """An ESI contact. Immutable."""

    class Category(str, Enum):
        """The category of a contact."""

        ALLIANCE = "alliance"
        CHARACTER = "character"
        CORPORATION = "corporation"
        FACTION = "faction"

        @classmethod
        def from_esi_contact_type(cls, contact_type) -> "EsiContact.Category":
            """Create from an ESI contact type."""
            mapper = {
                "alliance": cls.ALLIANCE,
                "character": cls.CHARACTER,
                "corporation": cls.CORPORATION,
                "faction": cls.FACTION,
            }
            return mapper[contact_type]

    contact_id: int
    contact_type: Category
    standing: float
    label_ids: FrozenSet[int] = field(default_factory=frozenset)
    is_war_target: bool = False

    def __post_init__(self):
        object.__setattr__(self, "contact_id", int(self.contact_id))
        object.__setattr__(self, "contact_type", self.Category(self.contact_type))
        object.__setattr__(self, "standing", float(self.standing))
        object.__setattr__(self, "label_ids", frozenset(self.label_ids))
        object.__setattr__(self, "is_war_targets", bool(self.is_war_target))

    def clone(self, **kwargs) -> "EsiContact":
        """Clone this object and optional overwrite field values with kwargs."""
        field_names = [field.name for field in fields(self.__class__)]
        params = {key: getattr(self, key) for key in field_names}
        params.update(kwargs)
        new_obj = self.__class__(**params)
        return new_obj

    def to_esi_dict(self) -> dict:
        """Return as a dict."""
        obj = {
            "contact_id": self.contact_id,
            "contact_type": self.Category(self.contact_type).value,
            "standing": self.standing,
            "is_war_target": self.is_war_target,
        }
        if self.label_ids:
            obj["label_ids"] = sorted(list(self.label_ids))
        return obj

    @classmethod
    def from_esi_dict(cls, esi_dict: dict) -> "EsiContact":
        """Create new objects from an ESI contact."""
        return cls(
            contact_id=esi_dict["contact_id"],
            contact_type=EsiContact.Category.from_esi_contact_type(
                esi_dict["contact_type"]
            ),
            standing=esi_dict["standing"],
            label_ids=esi_dict.get("label_ids") or frozenset(),
        )

    @classmethod
    def from_eve_entity(
        cls,
        eve_entity: EveEntity,
        standing: float,
        label_ids: Optional[Iterable[int]] = None,
        is_war_target: bool = False,
    ) -> "EsiContact":
        """Create new instance from an EveEntity object."""
        contact_type_map = {
            EveEntity.CATEGORY_ALLIANCE: cls.Category.ALLIANCE,
            EveEntity.CATEGORY_CHARACTER: cls.Category.CHARACTER,
            EveEntity.CATEGORY_CORPORATION: cls.Category.CORPORATION,
            EveEntity.CATEGORY_FACTION: cls.Category.FACTION,
        }
        if eve_entity.category not in contact_type_map:
            raise ValueError(
                f"{eve_entity}: Can not create from eve entity without category"
            )
        return cls(
            contact_id=eve_entity.id,
            contact_type=contact_type_map[eve_entity.category],
            is_war_target=is_war_target,
            label_ids=label_ids if label_ids else frozenset(),
            standing=standing,
        )

    @classmethod
    def from_eve_contact(cls, eve_contact: Any, label_ids=None) -> "EsiContact":
        """Create new instance from an EveContact object."""
        contact_type_map = {
            EveEntity.CATEGORY_ALLIANCE: cls.Category.ALLIANCE,
            EveEntity.CATEGORY_CHARACTER: cls.Category.CHARACTER,
            EveEntity.CATEGORY_CORPORATION: cls.Category.CORPORATION,
        }
        return cls(
            contact_id=eve_contact.eve_entity.id,
            contact_type=contact_type_map[eve_contact.eve_entity.category],
            standing=eve_contact.standing,
            label_ids=label_ids if label_ids else frozenset(),
        )


@dataclass
class _character:
    id: int
    corporation_id: int
    standing: float
    alliance_id: Optional[int] = None


@dataclass
class _corporation:
    id: int
    standing: float
    alliance_id: Optional[int] = None
    faction_id: Optional[int] = None
    is_war_target: bool = False


@dataclass
class _alliance:
    id: int
    standing: float
    is_war_target: bool = False


# pylint: disable = too-many-public-methods
@dataclass
class EsiContactsContainer:
    """Container of ESI contacts with their labels."""

    _contacts: Dict[int, EsiContact] = field(
        default_factory=dict, init=False, repr=False
    )
    _labels: Dict[int, EsiContactLabel] = field(
        default_factory=dict, init=False, repr=False
    )

    def add_eve_contacts(
        self, contacts: Iterable[object], label_ids: Optional[List[int]] = None
    ):
        """Add eve contacts to this container."""
        for contact in contacts:
            self.add_contact(EsiContact.from_eve_contact(contact, label_ids=label_ids))

    def add_contact(self, contact: EsiContact):
        """Add contact to container. Unknown label IDs will be removed."""
        if contact.label_ids:
            label_ids = {
                label_id for label_id in contact.label_ids if label_id in self._labels
            }
        else:
            label_ids = []
        self._contacts[contact.contact_id] = contact.clone(label_ids=label_ids)

    def add_label(self, label: EsiContactLabel):
        """Add contact label."""
        self._labels[label.id] = deepcopy(label)

    def clone(self) -> "EsiContactsContainer":
        """Return a clone of this object."""
        other = self.__class__.from_esi_contacts(
            contacts=self.contacts(), labels=self.labels()
        )
        return other

    def contact_by_id(self, contact_id: int) -> EsiContact:
        """Returns contact by it's ID.

        Raises ValueError when contact is not found.
        """
        try:
            return self._contacts[contact_id]
        except KeyError:
            raise ValueError(f"Contact with ID {contact_id} not found.") from None

    def contact_ids(self) -> Set[int]:
        """Return all contact IDs"""
        result = set(self._contacts)
        return result

    def contacts(self) -> Set[EsiContact]:
        """Fetch all contacts."""
        return set(self._contacts.values())

    # pylint: disable = protected-access
    def contacts_difference(
        self, other: "EsiContactsContainer"
    ) -> Tuple[Set[EsiContact], Set[EsiContact], Set[EsiContact]]:
        """Identify which contacts have been added, removed or changed."""
        current_contact_ids = set(self._contacts.keys())
        other_contact_ids = set(other._contacts.keys())
        removed = {
            contact
            for contact_id, contact in self._contacts.items()
            if contact_id in (current_contact_ids - other_contact_ids)
        }
        added = {
            contact
            for contact_id, contact in other._contacts.items()
            if contact_id in (other_contact_ids - current_contact_ids)
        }
        added_and_changed = set(other._contacts.values()) - set(self._contacts.values())
        changed = added_and_changed - added
        return added, removed, changed

    def contacts_to_esi_dicts(self) -> List[dict]:
        """Convert contacts into a stable dictionary."""
        return [
            obj.to_esi_dict()
            for obj in sorted(self._contacts.values(), key=lambda o: o.contact_id)
        ]

    def labels_to_esi_dicts(self) -> List[dict]:
        """Convert labels into a stable dictionary."""
        return [
            obj.to_esi_dict()
            for obj in sorted(self._labels.values(), key=lambda o: o.id)
        ]

    def label_by_id(self, label_id) -> EsiContactLabel:
        """Returns label by it's ID.

        Raises ValueError when label is not found.
        """
        try:
            return self._labels[label_id]
        except KeyError:
            raise ValueError(f"Label with ID {label_id} not found.") from None

    def labels(self) -> Set[EsiContactLabel]:
        """Fetch all labels."""
        return set(self._labels.values())

    def prune(self, unmask_war_targets=True, compress_contacts=False) -> int:
        """Prune contacts.

        When unmask_war_targets is True, prune will remove contacts shadowing
        the standing of war targets:
        - characters belonging to a war target with different standings
        - corporations belonging to a war target with different standings

        When compress_contacts is True, prune will also remove contacts
        which are unnecessary to calculate their effective standing:
        - faction contacts
        - characters and when their alliances exist as contact and has same standing
        - characters when their corporations exist as contact and has same standing
        - corporations when their alliances exist as contact and has same standing
        - neutral characters when their corporation and alliance do not exist as contact
        - neutral corporations when their corporation and alliance do not exist as contact
        - neutral alliances
        """
        # collect contacts
        characters = {
            x.contact_id: _character(
                id=x.contact_id, corporation_id=0, standing=x.standing
            )
            for x in self.contacts()
            if x.contact_type == EsiContact.Category.CHARACTER
        }
        corporations = {
            x.contact_id: _corporation(
                id=x.contact_id, standing=x.standing, is_war_target=x.is_war_target
            )
            for x in self.contacts()
            if x.contact_type == EsiContact.Category.CORPORATION
        }
        alliances = {
            x.contact_id: _alliance(
                id=x.contact_id, standing=x.standing, is_war_target=x.is_war_target
            )
            for x in self.contacts()
            if x.contact_type == EsiContact.Category.ALLIANCE
        }

        # add character affiliations
        for chunk in chunks(list(characters.keys()), 1000):
            affiliations = esi.client.Character.PostCharactersAffiliation(
                body=chunk
            ).result(use_etag=False)
            for x in affiliations:
                characters[x.character_id].corporation_id = x.corporation_id
                characters[x.character_id].alliance_id = x.alliance_id

        # add corporation affiliations
        for corporation_id in corporations.keys():
            info = esi.client.Corporation.GetCorporationsCorporationId(
                corporation_id=corporation_id
            ).result(use_etag=False)
            corporations[corporation_id].alliance_id = info.alliance_id

        # remove alliances
        if compress_contacts:
            for alliance_id in list(alliances.keys()):
                obj = alliances[alliance_id]
                if obj.standing == 0:
                    del alliances[alliance_id]

        # remove corporations
        for corporation_id in list(corporations.keys()):
            obj = corporations[corporation_id]
            alliance = None
            if obj.alliance_id:
                try:
                    alliance = alliances[obj.alliance_id]
                except KeyError:
                    pass

            if unmask_war_targets and alliance and alliance.is_war_target:
                del corporations[corporation_id]
                continue

            if compress_contacts and alliance and alliance.standing == obj.standing:
                del corporations[corporation_id]
                continue

            if compress_contacts and obj.standing == 0 and not alliance:
                del corporations[corporation_id]
                continue

        # remove characters
        for character_id in list(characters.keys()):
            obj = characters[character_id]
            try:
                corporation = corporations[obj.corporation_id]
            except KeyError:
                corporation = None

            if unmask_war_targets and corporation and corporation.is_war_target:
                del characters[character_id]
                continue

            if (
                compress_contacts
                and corporation
                and corporation.standing == obj.standing
            ):
                del characters[character_id]
                continue

            alliance = None
            if obj.alliance_id:
                try:
                    alliance = alliances[obj.alliance_id]
                except KeyError:
                    pass

            if unmask_war_targets and alliance and alliance.is_war_target:
                del characters[character_id]
                continue

            if compress_contacts and alliance and alliance.standing == obj.standing:
                del characters[character_id]
                continue

            if (
                compress_contacts
                and obj.standing == 0
                and not corporation
                and not alliance
            ):
                del characters[character_id]
                continue

        # updated contacts
        remaining_ids = characters.keys() | corporations.keys() | alliances.keys()
        if not compress_contacts:
            remaining_ids |= {
                x.contact_id
                for x in self.contacts()
                if x.contact_type == EsiContact.Category.FACTION
            }
        to_delete = self.contact_ids().difference(remaining_ids)
        for contact_id in to_delete:
            del self._contacts[contact_id]

        return len(to_delete)

    def remove_contact(self, contact: EsiContact):
        """Remove contact."""
        try:
            del self._contacts[contact.contact_id]
        except KeyError:
            raise ValueError(
                f"Unknown contact {contact} could not be removed."
            ) from None

    def remove_contacts(self, contacts: Iterable[EsiContact]):
        """Remove several contacts."""
        for contact in contacts:
            self.remove_contact(contact)

    def remove_war_targets(self):
        """Remove war targets."""
        self.remove_contacts(self.war_targets())

    def to_dict(self) -> dict:
        """Convert this object into a stable dictionary."""
        data = {
            "contacts": self.contacts_to_esi_dicts(),
            "labels": self.labels_to_esi_dicts(),
        }
        return data

    def update_contact(self, contact: EsiContact) -> bool:
        """Update an existing contact and return whether it was successful."""
        if contact.contact_id not in self._contacts:
            return False
        self._contacts[contact.contact_id] = contact
        return True

    def version_hash(self) -> str:
        """Calculate hash for current contacts & label in order to identify changes."""
        data = self.to_dict()
        return hashlib.md5(json.dumps(data).encode("utf-8")).hexdigest()

    def war_targets(self) -> Set[EsiContact]:
        """Fetch contacts that are war targets."""
        war_target_id = self.war_target_label_id()
        contacts = {obj for obj in self.contacts() if war_target_id in obj.label_ids}
        return contacts

    def war_target_label_id(self) -> Optional[int]:
        """Fetch the ID of the configured war target label."""
        for label in self._labels.values():
            if label.name.lower() == STANDINGSSYNC_WAR_TARGETS_LABEL_NAME.lower():
                return label.id
        return None

    @classmethod
    def from_esi_contacts(
        cls,
        contacts: Optional[Iterable[EsiContact]] = None,
        labels: Optional[Iterable[EsiContactLabel]] = None,
    ) -> "EsiContactsContainer":
        """Create new object from Esi contacts."""
        obj = cls()
        if labels:
            for label in labels:
                obj.add_label(label)
        if contacts:
            for contact in contacts:
                obj.add_contact(contact)
        return obj

    @classmethod
    def from_esi_dicts(
        cls,
        contacts: Optional[Iterable[dict]] = None,
        labels: Optional[Iterable[dict]] = None,
    ) -> "EsiContactsContainer":
        """Create new object from ESI contacts and labels."""
        obj = cls()
        if labels:
            for label in labels:
                obj.add_label(EsiContactLabel.from_esi_dict(label))
        if contacts:
            for contact in contacts:
                obj.add_contact(EsiContact.from_esi_dict(contact))
        return obj



"""Admin site for standingssync."""

# pylint: disable = missing-class-docstring, missing-function-docstring

from django.contrib import admin
from django.db.models import Prefetch
from django.utils.html import format_html
from eveuniverse.models import EveEntity

from standingssync import tasks
from standingssync.models import EveContact, EveWar, SyncedCharacter, SyncManager


@admin.register(EveContact)
class EveContactAdmin(admin.ModelAdmin):
    list_display = (
        "_entity_name",
        "_entity_category",
        "standing",
        "is_war_target",
        "manager",
    )
    list_display_links = None
    ordering = ("eve_entity__name",)
    list_select_related = True
    list_filter = ("manager", "eve_entity__category", "is_war_target")

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related(
            "eve_entity",
            "manager",
            "manager__alliance",
            "manager__character_ownership__character",
        )

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs) -> bool:
        return False

    @admin.display(ordering="eve_entity_id")
    def _entity_name(self, obj):
        return obj.eve_entity.name

    @admin.display(ordering="eve_entity__category")
    def _entity_category(self, obj):
        return obj.eve_entity.get_category_display()


class EveWarActiveWarsListFilter(admin.SimpleListFilter):
    title = "active wars"
    parameter_name = "active_wars"

    def lookups(self, request, model_admin):
        return (
            ("yes", "yes"),
            ("no", "no"),
        )

    def queryset(self, request, queryset):
        if self.value() == "yes":
            return queryset.filter(is_active=True)
        if self.value() == "no":
            return queryset.filter(is_active=False)
        return queryset


class EveWarEmptyWarsListFilter(admin.SimpleListFilter):
    title = "empty wars"
    parameter_name = "empty_wars"

    def lookups(self, request, model_admin):
        return (
            ("yes", "yes"),
            ("no", "no"),
        )

    def queryset(self, request, queryset):
        if self.value() == "yes":
            return queryset.empty()
        if self.value() == "no":
            return queryset.non_empty()
        return queryset


class EveWarStateListFilter(admin.SimpleListFilter):
    title = "state"
    parameter_name = "state"

    def lookups(self, request, model_admin):
        return EveWar.State.choices

    def queryset(self, request, queryset):
        if value := self.value():
            return queryset.filter(state=value)

        return queryset


class AlliesInline(admin.TabularInline):
    model = EveWar.allies.through


@admin.register(EveWar)
class EveWarAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "_state",
        "aggressor",
        "defender",
        "declared",
        "started",
        "retracted",
        "finished",
    )
    ordering = ("-id",)
    list_filter = (
        "declared",
        EveWarActiveWarsListFilter,
        EveWarStateListFilter,
        EveWarEmptyWarsListFilter,
    )
    search_fields = ("aggressor__name", "defender__name", "allies__name")
    inlines = (AlliesInline,)

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs) -> bool:
        return False

    def get_queryset(self, request):
        qs = super().get_queryset(request).annotate_state().annotate_is_active()
        return qs.prefetch_related(
            Prefetch("allies", queryset=EveEntity.objects.select_related())
        ).annotate_state()

    def _state(self, obj) -> str:
        return EveWar.State(obj.state).label

    # def _allies(self, obj):
    #     allies = sorted([str(ally) for ally in obj.allies.all()])
    #     return format_html("<br>".join(allies)) if allies else "-"

    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        p = round(EveWar.objects.updated_percentage() * 100, 1)
        extra_context["completion_p"] = p
        return super().changelist_view(request, extra_context=extra_context)


@admin.register(SyncedCharacter)
class SyncedCharacterAdmin(admin.ModelAdmin):
    list_display = (
        "_user",
        "_character_name",
        "_has_war_targets_label",
        "_is_fresh",
        "last_sync_at",
        "manager",
    )
    list_filter = (
        "manager",
        "last_sync_at",
        ("character_ownership__user", admin.RelatedOnlyFieldListFilter),
    )
    actions = ["sync_characters", "delete_all_contacts"]
    list_display_links = None

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related(
            "manager__alliance",
            "manager__character_ownership__character",
            "character_ownership__character",
            "character_ownership__user",
        )

    def has_add_permission(self, request):
        return False

    @admin.display(ordering="character_ownership__user")
    def _user(self, obj):
        return obj.character_ownership.user

    @admin.display(ordering="character_ownership__character__character_name")
    def _character_name(self, obj):
        return str(obj)

    @admin.display(boolean=True, description="Has WT label")
    def _has_war_targets_label(self, obj):
        return obj.has_war_targets_label

    @admin.display(boolean=True)
    def _is_fresh(self, obj):
        return obj.is_sync_fresh

    @admin.display(description="Start sync for selected synced characters")
    def sync_characters(self, request, queryset):
        names = []
        for obj in queryset:
            tasks.run_character_sync.delay(obj.pk)
            names.append(str(obj))
        names_text = ", ".join(names)
        self.message_user(request, f"Started normal syncing for: {names_text}")

    @admin.display(description="Delete all contacts of selected synced characters")
    def delete_all_contacts(self, request, queryset):
        names = []
        for obj in queryset:
            tasks.character_delete_all_contacts.delay(obj.pk)
            names.append(str(obj))
        names_text = ", ".join(names)
        self.message_user(request, f"Started deleting contacts for: {names_text}")


@admin.register(SyncManager)
class SyncManagerAdmin(admin.ModelAdmin):
    class Media:
        css = {"all": ("admin/custom_admin.css",)}

    list_display = (
        "_alliance_name",
        "_contacts_count",
        "_filters",
        "_characters_count",
        "_user",
        "_character_name",
        "_is_fresh",
        "last_sync_at",
    )
    list_filter = (("character_ownership__user", admin.RelatedOnlyFieldListFilter),)
    actions = ["start_sync_managers"]
    readonly_fields = (
        "alliance",
        "character_ownership",
        "version_hash",
        "last_sync_at",
    )
    fieldsets = (
        (
            None,
            {
                "fields": ("alliance", "character_ownership", "last_sync_at"),
            },
        ),
        (
            "Contact filters",
            {
                "fields": (
                    "compress_contacts",
                    "exclude_character_contacts",
                    "exclude_negative_standings",
                    "exclude_positive_standings",
                    "unmask_war_targets",
                ),
                "description": (
                    "Changes to filters will be applied at the next sync. "
                    "Filters do not effect war target contacts."
                ),
            },
        ),
    )

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("alliance", "character_ownership__character")

    def has_add_permission(self, request):
        return False

    @admin.display(ordering="character_ownership__user__username", description="User")
    def _user(self, obj):
        return obj.character_ownership.user if obj.character_ownership else None

    @admin.display(
        ordering="character_ownership__character__character_name",
        description="Character",
    )
    def _character_name(self, obj):
        try:
            return obj.character.character_name
        except ValueError:
            return ""

    @admin.display(ordering="alliance__alliance_name", description="Alliance")
    def _alliance_name(self, obj):
        return obj.alliance.alliance_name

    @admin.display(description="Contacts")
    def _contacts_count(self, obj):
        return f"{obj.contacts.count():,}"

    @admin.display(description="Filters")
    def _filters(self, obj):
        names = [
            "compress_contacts",
            "exclude_character_contacts",
            "exclude_negative_standings",
            "exclude_positive_standings",
            "unmask_war_targets",
        ]
        parts = []
        for n in names:
            value = getattr(obj, n)
            name_display = n.replace("_", " ")
            value_display = "<b>enabled</b>" if value else "disabled"
            parts.append(f"{name_display}: {value_display}")

        return format_html("<br>".join(parts))

    @admin.display(description="Characters")
    def _characters_count(self, obj):
        return f"{obj.synced_characters.count():,}"

    @admin.display(boolean=True)
    def _is_fresh(self, obj):
        return obj.is_sync_fresh

    @admin.display(description="Start sync for selected managers")
    def start_sync_managers(self, request, queryset):
        names = []
        for obj in queryset:
            tasks.run_manager_sync.delay(manager_pk=obj.pk, force_update=True)
            names.append(str(obj))
        text = f"Started syncing for: {', '.join(names)} "
        self.message_user(request, text)



from io import StringIO

from django.core.management import call_command
from django.core.management.base import CommandError

from app_utils.testing import NoSocketsTestCase

from standingssync.management.commands import standingssync_calc_war_ids
from standingssync.tests.factories import EveWarEmptyFactory, EveWarFactory


class TestCalcWarIds_Command(NoSocketsTestCase):
    def test_should_return_id_in_output(self):
        # given
        EveWarFactory(id=42)
        out = StringIO()

        # when
        call_command("standingssync_calc_war_ids", stdout=out)

        # then
        self.assertIn("42", out.getvalue())

    def test_should_allow_custom_number_of_special_ids(self):
        # given
        EveWarFactory(id=42)
        out = StringIO()

        # when
        call_command(
            "standingssync_calc_war_ids", "--max-special-ids", "30", stdout=out
        )

        # then
        self.assertIn("42", out.getvalue())

    def test_should_abort_when_no_wars_found(self):
        # when/then
        with self.assertRaises(CommandError):
            call_command("standingssync_calc_war_ids", stdout=StringIO())

    def test_should_abort_when_some_wars_are_empty(self):
        # given
        EveWarFactory()
        EveWarEmptyFactory()
        # when/then
        with self.assertRaises(CommandError):
            call_command("standingssync_calc_war_ids", stdout=StringIO())


class TestCalcWarIds_Logic(NoSocketsTestCase):
    def test_return_ids_and_min(self):
        # given
        EveWarFactory(id=1, is_finished=True)
        EveWarFactory(id=2)
        EveWarFactory(id=3)
        EveWarFactory(id=4, is_finished=True)
        EveWarFactory(id=5)
        EveWarFactory(id=6)

        # when
        got_ids, got_min = standingssync_calc_war_ids._calc_war_ids(2)

        # then
        self.assertCountEqual(got_ids, [2, 3])
        self.assertEqual(got_min, 5)

    def test_return_ids_and_min_2(self):
        # given
        EveWarFactory(id=1)

        # when
        got_ids, got_min = standingssync_calc_war_ids._calc_war_ids(2)

        # then
        self.assertCountEqual(got_ids, [1])
        self.assertEqual(got_min, 2)



from app_utils.testing import NoSocketsTestCase

from standingssync.tests.factories import (
    EveContactFactory,
    EveWarEmptyFactory,
    EveWarFactory,
    SyncedCharacterFactory,
    SyncManagerFactory,
)


class TestEveContactFactory(NoSocketsTestCase):
    def test_basic(self):
        x = EveContactFactory()
        self.assertEqual(x.standing, 5)
        self.assertFalse(x.is_war_target)

    def test_should_create_war_target(self):
        x = EveContactFactory(is_war_target=True)
        self.assertEqual(x.standing, -10)
        self.assertTrue(x.is_war_target)


class TestSyncedCharacterFactory(NoSocketsTestCase):
    def test_can_create_empty(self):
        sc = SyncedCharacterFactory()
        self.assertTrue(sc)

    def test_should_create_character_matching_the_manager(self):
        sm = SyncManagerFactory()
        sc = SyncedCharacterFactory(manager=sm)
        self.assertEqual(sc.character.alliance_id, sm.alliance.alliance_id)
        self.assertEqual(sc.character_ownership.user.character_ownerships.count(), 1)

    def test_should_create_al(self):
        sm = SyncManagerFactory()
        sc = SyncedCharacterFactory(manager=sm, create_alt=True)
        self.assertEqual(sc.character_ownership.user.character_ownerships.count(), 2)


class TestEveWarFactory(NoSocketsTestCase):
    def test_basic(self):
        war_1 = EveWarFactory()
        self.assertTrue(war_1.id)
        war_2 = EveWarEmptyFactory()
        self.assertTrue(war_2.id)



from http import HTTPStatus
from unittest.mock import patch

import pook

from eveuniverse.tests.testdata.factories_2 import (
    EveEntityAllianceFactory,
    EveEntityCharacterFactory,
    EveEntityCorporationFactory,
    EveEntityFactionFactory,
)

from app_utils.testing import NoSocketsTestCase

from standingssync.core.esi_contacts import (
    EsiContact,
    EsiContactLabel,
    EsiContactsContainer,
)
from standingssync.tests.factories import (
    EsiContactAllianceFactory,
    EsiContactCharacterFactory,
    EsiContactCorporationFactory,
    EsiContactFactionFactory,
    EsiContactFactory,
    EsiContactLabelFactory,
    EveContactFactory,
    make_esi_url,
)

MODULE_PATH = "standingssync.core.esi_contacts"
WAR_TARGET_LABEL = "WAR TARGETS"


class TestEsiContact(NoSocketsTestCase):
    def test_should_create_new_1(self):
        # when
        obj = EsiContact(1001, EsiContact.Category.CHARACTER, 5.0)
        # then
        self.assertEqual(obj.contact_id, 1001)
        self.assertEqual(obj.contact_type, EsiContact.Category.CHARACTER)
        self.assertEqual(obj.standing, 5.0)

    def test_should_create_new_2(self):
        # when
        obj = EsiContact(1001, "character", 5.0)  # type: ignore
        # then
        self.assertEqual(obj.contact_id, 1001)
        self.assertEqual(obj.contact_type, EsiContact.Category.CHARACTER)
        self.assertEqual(obj.standing, 5.0)

    def test_should_create_new_3(self):
        # when/then
        with self.assertRaises(ValueError):
            EsiContact(1001, "xyz", 5.0)  # type: ignore

    def test_should_create_all_types(self):
        # given
        params = [
            ("character", EsiContact.Category.CHARACTER),
            ("corporation", EsiContact.Category.CORPORATION),
            ("alliance", EsiContact.Category.ALLIANCE),
            ("faction", EsiContact.Category.FACTION),
        ]
        for input, expected in params:
            with self.subTest(input=input):
                # when
                obj = EsiContact(1001, input, 5.0)
                # then
                self.assertEqual(obj.contact_id, 1001)
                self.assertEqual(obj.contact_type, expected)
                self.assertEqual(obj.standing, 5.0)

    def test_should_clone_contact_1(self):
        # given
        a = EsiContact(1, EsiContact.Category.CHARACTER, 5.0, frozenset([1, 2]))
        # when
        b = a.clone()
        # then
        self.assertEqual(a, b)

    def test_should_clone_contact_2(self):
        # given
        a = EsiContact(1, EsiContact.Category.CHARACTER, 5.0, frozenset([1, 2]))
        # when
        b = a.clone(standing=-10)
        # then
        self.assertEqual(b.contact_id, a.contact_id)
        self.assertEqual(b.contact_type, a.contact_type)
        self.assertEqual(b.label_ids, a.label_ids)
        self.assertEqual(b.standing, -10)

    def test_should_create_from_esi_character(self):
        # given
        esi_dict = {"contact_id": 1, "contact_type": "character", "standing": 5.0}
        # when
        obj = EsiContact.from_esi_dict(esi_dict)
        # then
        self.assertEqual(obj, EsiContact(1, EsiContact.Category.CHARACTER, 5.0))

    def test_should_create_from_esi_corporation(self):
        # given
        esi_dict = {"contact_id": 1, "contact_type": "corporation", "standing": 5.0}
        # when
        obj = EsiContact.from_esi_dict(esi_dict)
        # then
        self.assertEqual(obj, EsiContact(1, EsiContact.Category.CORPORATION, 5.0))

    def test_should_create_from_esi_alliance(self):
        # given
        esi_dict = {"contact_id": 1, "contact_type": "alliance", "standing": 5.0}
        # when
        obj = EsiContact.from_esi_dict(esi_dict)
        # then
        self.assertEqual(obj, EsiContact(1, EsiContact.Category.ALLIANCE, 5.0))

    def test_should_create_from_esi_contact_when_labels_are_none(self):
        # given
        esi_dict = {
            "contact_id": 1,
            "contact_type": "alliance",
            "standing": 5.0,
            "label_ids": None,
        }
        # when
        obj = EsiContact.from_esi_dict(esi_dict)
        # then
        self.assertEqual(obj, EsiContact(1, EsiContact.Category.ALLIANCE, 5.0))

    def test_should_create_from_esi_faction(self):
        # given
        esi_dict = {"contact_id": 1, "contact_type": "faction", "standing": 5.0}
        # when
        obj = EsiContact.from_esi_dict(esi_dict)
        # then
        self.assertEqual(obj, EsiContact(1, EsiContact.Category.FACTION, 5.0))

    def test_should_create_from_eve_entities(self):
        # given
        params = [
            (EveEntityCharacterFactory(), EsiContact.Category.CHARACTER),
            (EveEntityCorporationFactory(), EsiContact.Category.CORPORATION),
            (EveEntityAllianceFactory(), EsiContact.Category.ALLIANCE),
            (EveEntityFactionFactory(), EsiContact.Category.FACTION),
        ]
        for eve_entity, expected in params:
            with self.subTest(category=eve_entity.category):
                # when
                got = EsiContact.from_eve_entity(eve_entity=eve_entity, standing=5.0)
                # then
                self.assertEqual(got, EsiContact(eve_entity.id, expected, 5.0))
                self.assertFalse(got.is_war_target)

    def test_should_create_from_eve_entities_as_war_target(self):
        # when
        got = EsiContact.from_eve_entity(
            eve_entity=EveEntityCharacterFactory(), standing=-10, is_war_target=True
        )
        # then
        self.assertTrue(got.is_war_target)


class TestEsiContactLabel(NoSocketsTestCase):
    def test_should_create_from_esi_dict(self):
        # given
        esi_dict = {"label_id": 42, "label_name": "alpha"}
        # when
        result = EsiContactLabel.from_esi_dict(esi_dict)
        # then
        self.assertEqual(result.id, 42)
        self.assertEqual(result.name, "alpha")

    def test_should_convert_to_dict(self):
        # given
        obj = EsiContactLabel(42, "alpha")
        # when
        result = obj.to_dict()
        # then
        expected = {42: "alpha"}
        self.assertDictEqual(expected, result)

    def test_should_convert_to_esi_dict(self):
        # given
        obj = EsiContactLabel(42, "alpha")
        # when
        result = obj.to_esi_dict()
        # then
        expected = {"label_id": 42, "label_name": "alpha"}
        self.assertDictEqual(expected, result)


@patch(MODULE_PATH + ".STANDINGSSYNC_WAR_TARGETS_LABEL_NAME", WAR_TARGET_LABEL)
class TestEsiContactsContainer(NoSocketsTestCase):
    def test_should_create_empty(self):
        # when
        obj = EsiContactsContainer()
        # then
        self.assertIsInstance(obj, EsiContactsContainer)

    def test_should_create_from_contacts_dict(self):
        # given
        contact_1 = EsiContactFactory()
        contact_2 = EsiContactFactory()
        esi_contacts = [contact_1.to_esi_dict(), contact_2.to_esi_dict()]
        # when
        obj = EsiContactsContainer.from_esi_dicts(esi_contacts)
        # then
        expected = {contact_1, contact_2}
        self.assertSetEqual(obj.contacts(), expected)

    def test_should_create_from_contacts_dict_w_labels(self):
        # given
        label_1 = EsiContactLabelFactory()
        contact_1 = EsiContactFactory(label_ids=[label_1.id])
        label_2 = EsiContactLabelFactory()
        contact_2 = EsiContactFactory(label_ids=[label_1.id, label_2.id])
        esi_contacts = [contact_1.to_esi_dict(), contact_2.to_esi_dict()]
        esi_labels = [label_1.to_esi_dict(), label_2.to_esi_dict()]
        # when
        obj = EsiContactsContainer.from_esi_dicts(esi_contacts, esi_labels)
        # then
        expected = {contact_1, contact_2}
        self.assertSetEqual(obj.contacts(), expected)

    def test_should_create_from_dicts_labels_only(self):
        # given
        label = EsiContactLabelFactory()
        esi_labels = [label.to_esi_dict()]
        # when
        obj = EsiContactsContainer.from_esi_dicts(labels=esi_labels)
        # then
        self.assertEqual(obj.labels(), {label})

    def test_should_create_empty_for_dicts(self):
        # when
        obj = EsiContactsContainer.from_esi_dicts()
        # then
        self.assertIsInstance(obj, EsiContactsContainer)

    def test_should_create_from_esi_contacts(self):
        # given
        contact_1 = EsiContactFactory()
        contact_2 = EsiContactFactory()
        esi_contacts = [contact_1, contact_2]
        # when
        obj = EsiContactsContainer.from_esi_contacts(esi_contacts)
        # then
        expected = {contact_1, contact_2}
        self.assertSetEqual(obj.contacts(), expected)

    def test_should_create_from_esi_contacts_w_labels(self):
        # given
        label_1 = EsiContactLabelFactory()
        contact_1 = EsiContactFactory(label_ids=[label_1.id])
        label_2 = EsiContactLabelFactory()
        contact_2 = EsiContactFactory(label_ids=[label_1.id, label_2.id])
        esi_contacts = [contact_1, contact_2]
        esi_labels = [label_1, label_2]
        # when
        obj = EsiContactsContainer.from_esi_contacts(esi_contacts, esi_labels)
        # then
        expected = {contact_1, contact_2}
        self.assertSetEqual(obj.contacts(), expected)

    def test_should_return_contacts_by_id(self):
        # given
        c1 = EsiContactFactory()
        c2 = EsiContactFactory()
        contacts = EsiContactsContainer.from_esi_contacts([c1, c2])
        # when
        result = contacts.contact_by_id(c1.contact_id)
        # then
        self.assertEqual(result, c1)

    def test_should_raise_error_when_contact_not_found(self):
        # given
        c1 = EsiContactFactory()
        c2 = EsiContactFactory()
        contacts = EsiContactsContainer.from_esi_contacts([c1, c2])
        # when/then
        with self.assertRaises(ValueError):
            contacts.contact_by_id(123456)

    def test_should_raise_error_when_label_not_found(self):
        # given
        contacts = EsiContactsContainer.from_esi_contacts()
        # when/then
        with self.assertRaises(ValueError):
            contacts.label_by_id(123456)

    def test_should_find_label_by_id(self):
        # given
        label = EsiContactLabelFactory()
        contacts = EsiContactsContainer.from_esi_contacts(labels=[label])
        # when
        new_label = contacts.label_by_id(label.id)
        # then
        self.assertEqual(label, new_label)

    def test_should_remove_unknown_label_ids(self):
        # given
        label_1 = EsiContactLabelFactory()
        label_2 = EsiContactLabelFactory()
        contacts = EsiContactsContainer.from_esi_contacts(labels=[label_1])
        contact_1 = EsiContactFactory(label_ids=[label_1.id, label_2.id])
        # when
        contacts.add_contact(contact_1)
        # then
        contact_1a = contacts.contact_by_id(contact_1.contact_id)
        self.assertEqual(contact_1a.label_ids, frozenset([label_1.id]))

    def test_should_return_labels(self):
        # given
        labels = {EsiContactLabelFactory(), EsiContactLabelFactory()}
        obj = EsiContactsContainer.from_esi_contacts(labels=labels)
        # when/then
        self.assertSetEqual(labels, obj.labels())

    def test_should_return_contacts(self):
        # given
        contacts = {EsiContactFactory(), EsiContactFactory()}
        obj = EsiContactsContainer.from_esi_contacts(contacts)
        # when/then
        self.assertSetEqual(contacts, obj.contacts())

    def test_should_remove_contact(self):
        # given
        contact_1 = EsiContactFactory()
        contact_2 = EsiContactFactory()
        esi_contacts = [contact_1.to_esi_dict(), contact_2.to_esi_dict()]
        obj = EsiContactsContainer.from_esi_dicts(esi_contacts)
        # when
        obj.remove_contact(contact_2)
        # then
        expected = {contact_1}
        self.assertSetEqual(obj.contacts(), expected)

    def test_should_remove_several_contacts(self):
        # given
        contact_1 = EsiContactFactory()
        contact_2 = EsiContactFactory()
        contact_3 = EsiContactFactory()
        obj = EsiContactsContainer.from_esi_contacts([contact_1, contact_2, contact_3])
        # when
        obj.remove_contacts([contact_1, contact_2])
        # then
        expected = {contact_3}
        self.assertSetEqual(obj.contacts(), expected)

    def test_should_raise_error_when_trying_to_remove_unknown_contact(self):
        # given
        contact_1 = EsiContactFactory()
        contact_2 = EsiContactFactory()
        obj = EsiContactsContainer.from_esi_contacts([contact_1])
        # when/then
        with self.assertRaises(ValueError):
            obj.remove_contact(contact_2)

    def test_should_convert_to_esi_dict(self):
        # given
        label_1 = EsiContactLabelFactory(id=1)
        contact_1 = EsiContactFactory(contact_id=11, label_ids=[label_1.id])
        label_2 = EsiContactLabelFactory(id=2)
        contact_2 = EsiContactFactory(contact_id=12, label_ids=[label_1.id, label_2.id])
        esi_contacts = [contact_1.to_esi_dict(), contact_2.to_esi_dict()]
        esi_labels = [label_1.to_esi_dict(), label_2.to_esi_dict()]
        obj = EsiContactsContainer.from_esi_dicts(esi_contacts, esi_labels)
        # when/then
        self.assertListEqual(obj.contacts_to_esi_dicts(), esi_contacts)
        self.assertListEqual(obj.labels_to_esi_dicts(), esi_labels)

    def test_should_generate_version_hash(self):
        # given
        label_1 = EsiContactLabelFactory()
        contact_1 = EsiContactFactory(label_ids=[label_1.id])
        label_2 = EsiContactLabelFactory()
        contact_2 = EsiContactFactory(label_ids=[label_1.id, label_2.id])
        esi_contacts = [contact_1.to_esi_dict(), contact_2.to_esi_dict()]
        esi_labels = [label_1.to_esi_dict(), label_2.to_esi_dict()]
        obj_1 = EsiContactsContainer.from_esi_dicts(esi_contacts, esi_labels)
        obj_2 = EsiContactsContainer.from_esi_dicts(esi_contacts, esi_labels)
        # when/then
        self.assertEqual(obj_1.version_hash(), obj_2.version_hash())

    def test_should_find_war_target_id(self):
        # given
        label_1 = EsiContactLabelFactory(name=WAR_TARGET_LABEL)
        label_2 = EsiContactLabelFactory()
        obj = EsiContactsContainer.from_esi_contacts(labels=[label_1, label_2])
        # when
        result = obj.war_target_label_id()
        # then
        self.assertEqual(result, label_1.id)

    def test_should_not_find_war_target_id(self):
        # given
        label_1 = EsiContactLabelFactory(name="alpha")
        label_2 = EsiContactLabelFactory(name="bravo")
        obj = EsiContactsContainer.from_esi_contacts(labels=[label_1, label_2])
        # when
        result = obj.war_target_label_id()
        # then
        self.assertIsNone(result)

    def test_should_return_war_targets(self):
        # given
        wt_label = EsiContactLabelFactory(name=WAR_TARGET_LABEL)
        other_label = EsiContactLabelFactory()
        other_contact = EsiContactFactory(label_ids=[other_label.id])
        war_target = EsiContactFactory(label_ids=[wt_label.id, other_label.id])
        obj = EsiContactsContainer.from_esi_contacts(
            contacts=[other_contact, war_target], labels=[wt_label, other_label]
        )
        # when
        result = obj.war_targets()
        # then
        self.assertSetEqual(result, {war_target})

    def test_should_remove_war_targets(self):
        # given
        wt_label = EsiContactLabelFactory(name=WAR_TARGET_LABEL)
        other_label = EsiContactLabelFactory()
        other_contact = EsiContactFactory(label_ids=[other_label.id])
        war_target = EsiContactFactory(label_ids=[wt_label.id, other_label.id])
        obj = EsiContactsContainer.from_esi_contacts(
            contacts=[other_contact, war_target], labels=[wt_label, other_label]
        )
        # when
        obj.remove_war_targets()
        # then
        self.assertSetEqual(obj.contacts(), {other_contact})

    def test_should_add_eve_contacts(self):
        # given
        obj = EsiContactsContainer()
        contact_1 = EveContactFactory()
        contact_2 = EveContactFactory()
        # when
        obj.add_eve_contacts([contact_1, contact_2])
        # then
        expected = {
            EsiContact.from_eve_contact(contact_1),
            EsiContact.from_eve_contact(contact_2),
        }
        self.assertSetEqual(obj.contacts(), expected)

    def test_should_add_eve_contacts_w_labels(self):
        # given
        label = EsiContactLabelFactory()
        obj = EsiContactsContainer()
        obj.add_label(label)
        label_ids = [label.id]
        contact_1 = EveContactFactory()
        contact_2 = EveContactFactory()
        # when
        obj.add_eve_contacts([contact_1, contact_2], label_ids=label_ids)
        # then
        expected = {
            EsiContact.from_eve_contact(contact_1, label_ids=label_ids),
            EsiContact.from_eve_contact(contact_2, label_ids),
        }
        self.assertSetEqual(obj.contacts(), expected)

    def test_should_create_clone(self):
        # given
        label_1 = EsiContactLabelFactory()
        label_2 = EsiContactLabelFactory()
        contact_1 = EsiContactFactory(label_ids=[label_1.id])
        contact_2 = EsiContactFactory(label_ids=[label_2.id])
        first = EsiContactsContainer.from_esi_contacts(
            contacts=[contact_1, contact_2], labels=[label_1, label_2]
        )
        # when
        second = first.clone()
        # then
        self.assertEqual(first, second)
        second.remove_contact(contact_1)
        self.assertNotIn(contact_1, second.contacts())
        self.assertIn(contact_1, first.contacts())

    def test_should_return_all_contact_ids(self):
        # given
        c1 = EsiContactFactory()
        c2 = EsiContactFactory()
        contacts = EsiContactsContainer.from_esi_contacts([c1, c2])
        # when
        result = contacts.contact_ids()
        # then
        self.assertSetEqual(result, {c1.contact_id, c2.contact_id})


class TestEsiContactsCloneComparisons(NoSocketsTestCase):
    def test_should_return_contacts_difference(self):
        # given
        c1 = EsiContactFactory()
        c2 = EsiContactFactory()
        c3 = EsiContactFactory()
        c4 = EsiContactFactory(standing=5)
        c4a = c4.clone(standing=-10)
        a = EsiContactsContainer.from_esi_contacts([c1, c2, c4])
        b = EsiContactsContainer.from_esi_contacts([c1, c3, c4a])
        # when
        added, removed, changed = a.contacts_difference(b)
        # then
        self.assertSetEqual(added, {c3})
        self.assertSetEqual(removed, {c2})
        self.assertSetEqual(changed, {c4a})


class TestEsiContacts_Prune(NoSocketsTestCase):
    @pook.on
    def test_should_remove_neutral_alliances(
        self,
    ):
        # given
        alliance = EsiContactAllianceFactory(standing=0)
        contacts = EsiContactsContainer.from_esi_contacts([alliance])

        # when
        contacts.prune(unmask_war_targets=True, compress_contacts=True)

        # then
        got = contacts.contacts()
        want = set()
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_not_remove_neutral_alliances(
        self,
    ):
        # given
        alliance = EsiContactAllianceFactory(standing=0)
        contacts = EsiContactsContainer.from_esi_contacts([alliance])

        # when
        contacts.prune(unmask_war_targets=True, compress_contacts=False)

        # then
        got = contacts.contacts()
        want = {alliance}
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_remove_faction_contacts(
        self,
    ):
        # given
        faction = EsiContactFactionFactory(standing=5)
        contacts = EsiContactsContainer.from_esi_contacts([faction])

        # when
        contacts.prune(unmask_war_targets=True, compress_contacts=True)

        # then
        got = contacts.contacts()
        want = set()
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_not_remove_faction_contacts(
        self,
    ):
        # given
        faction = EsiContactFactionFactory(standing=5)
        contacts = EsiContactsContainer.from_esi_contacts([faction])

        # when
        contacts.prune(unmask_war_targets=True, compress_contacts=False)

        # then
        got = contacts.contacts()
        want = {faction}
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_remove_character_when_corporation_exists_and_has_same_standing(
        self,
    ):
        # given
        corporation = EsiContactCorporationFactory(contact_id=2001, standing=-5)
        character_1 = EsiContactCharacterFactory(standing=-5)
        character_2 = EsiContactCharacterFactory(standing=5)
        pook.post(
            make_esi_url("characters/affiliation"),
            reply=HTTPStatus.OK,
            response_json=[
                {
                    "character_id": character_1.contact_id,
                    "corporation_id": corporation.contact_id,
                },
                {
                    "character_id": character_2.contact_id,
                    "corporation_id": corporation.contact_id,
                },
            ],
        )
        pook.get(
            make_esi_url(f"corporations/{corporation.contact_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "ceo_id": 90000001,
                "creator_id": 90000001,
                "member_count": 42,
                "name": "name",
                "tax_rate": 0,
                "ticker": "ticker",
            },
        )
        contacts = EsiContactsContainer.from_esi_contacts(
            [corporation, character_1, character_2]
        )

        # when
        contacts.prune(unmask_war_targets=True, compress_contacts=True)

        # then
        got = contacts.contacts()
        want = {corporation, character_2}
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_remove_character_when_alliance_exists_and_has_same_standing(
        self,
    ):
        # given
        corporation_id = 2001
        alliance = EsiContactAllianceFactory(contact_id=3001, standing=-5)
        character_1 = EsiContactCharacterFactory(contact_id=1001, standing=-5)
        character_2 = EsiContactCharacterFactory(contact_id=1002, standing=5)
        pook.post(
            make_esi_url("characters/affiliation"),
            reply=HTTPStatus.OK,
            response_json=[
                {
                    "alliance_id": alliance.contact_id,
                    "character_id": character_1.contact_id,
                    "corporation_id": corporation_id,
                },
                {
                    "character_id": character_2.contact_id,
                    "corporation_id": corporation_id,
                },
            ],
        )
        contacts = EsiContactsContainer.from_esi_contacts(
            [alliance, character_1, character_2]
        )

        # when
        contacts.prune(unmask_war_targets=True, compress_contacts=True)

        # then
        got = contacts.contacts()
        want = {alliance, character_2}
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_remove_corporation_when_alliance_exists_and_has_same_standing(
        self,
    ):
        # given
        corporation_1_id = 2001
        corporation_2_id = 2002
        alliance = EsiContactAllianceFactory(contact_id=3001, standing=-5)
        corporation_1 = EsiContactCorporationFactory(
            contact_id=corporation_1_id, standing=-5
        )
        corporation_2 = EsiContactCorporationFactory(
            contact_id=corporation_2_id, standing=5
        )
        pook.get(
            make_esi_url(f"corporations/{corporation_1_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "ceo_id": 90000001,
                "creator_id": 90000001,
                "alliance_id": alliance.contact_id,
                "member_count": 42,
                "name": "name",
                "tax_rate": 0,
                "ticker": "ticker",
            },
            persist=True,
        )
        pook.get(
            make_esi_url(f"corporations/{corporation_2_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "ceo_id": 90000001,
                "creator_id": 90000001,
                "alliance_id": alliance.contact_id,
                "member_count": 42,
                "name": "name",
                "tax_rate": 0,
                "ticker": "ticker",
            },
            persist=True,
        )
        all_contacts = [alliance, corporation_1, corporation_2]
        contacts_1 = EsiContactsContainer.from_esi_contacts(all_contacts)
        contacts_2 = EsiContactsContainer.from_esi_contacts(all_contacts)

        # when
        contacts_1.prune(unmask_war_targets=True, compress_contacts=True)
        contacts_2.prune(unmask_war_targets=True, compress_contacts=False)

        # then
        got = contacts_1.contacts()
        want = {alliance, corporation_2}
        self.assertSetEqual(got, want)
        self.assertSetEqual(contacts_2.contacts(), set(all_contacts))

    @pook.on
    def test_should_remove_neutral_characters_when_their_corporation_and_alliance_are_not_contacts(
        self,
    ):
        # given
        corporation = EsiContactCorporationFactory(contact_id=2001, standing=-5)
        character_1 = EsiContactCharacterFactory(contact_id=1001, standing=0)
        character_2 = EsiContactCharacterFactory(contact_id=1011, standing=0)
        pook.post(
            make_esi_url("characters/affiliation"),
            reply=HTTPStatus.OK,
            response_json=[
                {
                    "character_id": character_1.contact_id,
                    "corporation_id": corporation.contact_id,
                },
                {
                    "character_id": character_2.contact_id,
                    "corporation_id": 2011,
                },
            ],
            persist=True,
        )
        pook.get(
            make_esi_url(f"corporations/{corporation.contact_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "ceo_id": 90000001,
                "creator_id": 90000001,
                "member_count": 42,
                "name": "name",
                "tax_rate": 0,
                "ticker": "ticker",
            },
            persist=True,
        )
        all_contacts = [corporation, character_1, character_2]
        contacts_1 = EsiContactsContainer.from_esi_contacts(all_contacts)
        contacts_2 = EsiContactsContainer.from_esi_contacts(all_contacts)

        # when
        contacts_1.prune(unmask_war_targets=True, compress_contacts=True)
        contacts_2.prune(unmask_war_targets=True, compress_contacts=False)

        # then
        got = contacts_1.contacts()
        want = {corporation, character_1}
        self.assertSetEqual(got, want)
        self.assertSetEqual(contacts_2.contacts(), set(all_contacts))

    @pook.on
    def test_should_remove_neutral_corporations_when_their_alliances_are_not_contacts(
        self,
    ):
        # given
        alliance = EsiContactAllianceFactory(contact_id=3001, standing=-5)
        corporation_1 = EsiContactCorporationFactory(contact_id=2001, standing=0)
        corporation_2 = EsiContactCorporationFactory(contact_id=2002, standing=0)
        pook.get(
            make_esi_url(f"corporations/{corporation_1.contact_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "ceo_id": 90000001,
                "creator_id": 90000001,
                "alliance_id": alliance.contact_id,
                "member_count": 42,
                "name": "name",
                "tax_rate": 0,
                "ticker": "ticker",
            },
            persist=True,
        )
        pook.get(
            make_esi_url(f"corporations/{corporation_2.contact_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "ceo_id": 90000001,
                "creator_id": 90000001,
                "member_count": 42,
                "name": "name",
                "tax_rate": 0,
                "ticker": "ticker",
            },
            persist=True,
        )
        all_contacts = [alliance, corporation_1, corporation_2]
        contacts_1 = EsiContactsContainer.from_esi_contacts(all_contacts)
        contacts_2 = EsiContactsContainer.from_esi_contacts(all_contacts)

        # when
        contacts_1.prune(unmask_war_targets=True, compress_contacts=True)
        contacts_2.prune(unmask_war_targets=True, compress_contacts=False)

        # then
        got = contacts_1.contacts()

        want = {alliance, corporation_1}
        self.assertSetEqual(got, want)
        self.assertSetEqual(contacts_2.contacts(), set(all_contacts))

    @pook.on
    def test_should_remove_corporations_when_they_shadow_a_war_target(
        self,
    ):
        # given
        alliance = EsiContactAllianceFactory(
            contact_id=3001, standing=-10, is_war_target=True
        )
        corporation = EsiContactCorporationFactory(contact_id=2001, standing=0)
        pook.get(
            make_esi_url(f"corporations/{corporation.contact_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "ceo_id": 90000001,
                "creator_id": 90000001,
                "alliance_id": alliance.contact_id,
                "member_count": 42,
                "name": "name",
                "tax_rate": 0,
                "ticker": "ticker",
            },
        )
        contacts = EsiContactsContainer.from_esi_contacts([alliance, corporation])

        # when
        contacts.prune(unmask_war_targets=True, compress_contacts=True)

        # then
        got = contacts.contacts()

        want = {alliance}
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_not_remove_corporations_when_they_shadow_a_war_target(
        self,
    ):
        # given
        alliance = EsiContactAllianceFactory(
            contact_id=3001, standing=-10, is_war_target=True
        )
        corporation = EsiContactCorporationFactory(contact_id=2001, standing=0)
        pook.get(
            make_esi_url(f"corporations/{corporation.contact_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "ceo_id": 90000001,
                "creator_id": 90000001,
                "alliance_id": alliance.contact_id,
                "member_count": 42,
                "name": "name",
                "tax_rate": 0,
                "ticker": "ticker",
            },
        )
        contacts = EsiContactsContainer.from_esi_contacts([alliance, corporation])

        # when
        contacts.prune(unmask_war_targets=False, compress_contacts=True)

        # then
        got = contacts.contacts()

        want = {alliance, corporation}
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_remove_characters_when_they_shadow_a_war_target_corporation(
        self,
    ):
        # given
        corporation = EsiContactCorporationFactory(
            contact_id=2001, standing=-10, is_war_target=True
        )
        character = EsiContactCharacterFactory(standing=5)
        pook.post(
            make_esi_url("characters/affiliation"),
            reply=HTTPStatus.OK,
            response_json=[
                {
                    "character_id": character.contact_id,
                    "corporation_id": corporation.contact_id,
                },
            ],
        )
        pook.get(
            make_esi_url(f"corporations/{corporation.contact_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "ceo_id": 90000001,
                "creator_id": 90000001,
                "member_count": 42,
                "name": "name",
                "tax_rate": 0,
                "ticker": "ticker",
            },
        )
        contacts = EsiContactsContainer.from_esi_contacts([corporation, character])

        # when
        contacts.prune(unmask_war_targets=True, compress_contacts=True)

        # then
        got = contacts.contacts()
        want = {corporation}
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_not_remove_characters_when_they_shadow_a_war_target_corporation(
        self,
    ):
        # given
        corporation = EsiContactCorporationFactory(
            contact_id=2001, standing=-10, is_war_target=True
        )
        character = EsiContactCharacterFactory(standing=5)
        pook.post(
            make_esi_url("characters/affiliation"),
            reply=HTTPStatus.OK,
            response_json=[
                {
                    "character_id": character.contact_id,
                    "corporation_id": corporation.contact_id,
                },
            ],
        )
        pook.get(
            make_esi_url(f"corporations/{corporation.contact_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "ceo_id": 90000001,
                "creator_id": 90000001,
                "member_count": 42,
                "name": "name",
                "tax_rate": 0,
                "ticker": "ticker",
            },
        )
        contacts = EsiContactsContainer.from_esi_contacts([corporation, character])

        # when
        contacts.prune(unmask_war_targets=False, compress_contacts=True)

        # then
        got = contacts.contacts()
        want = {corporation, character}
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_remove_characters_when_they_shadow_a_war_target_alliance(
        self,
    ):
        # given
        alliance = EsiContactAllianceFactory(
            contact_id=3001, standing=-10, is_war_target=True
        )
        character = EsiContactCharacterFactory(standing=5)
        pook.post(
            make_esi_url("characters/affiliation"),
            reply=HTTPStatus.OK,
            response_json=[
                {
                    "alliance_id": alliance.contact_id,
                    "character_id": character.contact_id,
                    "corporation_id": 2001,
                },
            ],
        )
        contacts = EsiContactsContainer.from_esi_contacts([alliance, character])

        # when
        contacts.prune(unmask_war_targets=True, compress_contacts=True)

        # then
        got = contacts.contacts()
        want = {alliance}
        self.assertSetEqual(got, want)

    @pook.on
    def test_should_not_remove_characters_when_they_shadow_a_war_target_alliance(
        self,
    ):
        # given
        alliance = EsiContactAllianceFactory(
            contact_id=3001, standing=-10, is_war_target=True
        )
        character = EsiContactCharacterFactory(standing=5)
        pook.post(
            make_esi_url("characters/affiliation"),
            reply=HTTPStatus.OK,
            response_json=[
                {
                    "alliance_id": alliance.contact_id,
                    "character_id": character.contact_id,
                    "corporation_id": 2001,
                },
            ],
        )
        contacts = EsiContactsContainer.from_esi_contacts([alliance, character])

        # when
        contacts.prune(unmask_war_targets=False, compress_contacts=True)

        # then
        got = contacts.contacts()
        want = {alliance, character}
        self.assertSetEqual(got, want)



from http import HTTPStatus

from django.contrib.admin.sites import AdminSite
from django.test import TestCase

from app_utils.testdata_factories import UserFactory

from standingssync.admin import SyncManagerAdmin
from standingssync.models import SyncManager
from standingssync.tests.factories import (
    EveContactFactory,
    SyncedCharacterFactory,
    SyncManagerFactory,
)


class TestSyncedCharacterChangeList(TestCase):
    def test_should_open_main_page(self):
        # given
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        # when
        response = self.client.get("/admin/standingssync/syncedcharacter/")
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_should_filter_by_character(self):
        # given
        SyncedCharacterFactory()
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        # when
        response = self.client.get("/admin/standingssync/syncedcharacter/?o=2")
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)


class TestSyncManagerChangeList_HTTP(TestCase):
    def test_can_open_page_normally(self):
        # given
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        SyncManagerFactory()

        # when
        response = self.client.get("/admin/standingssync/syncmanager/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)


class TestSyncManagerChangeList_Admin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = SyncManagerAdmin(model=SyncManager, admin_site=AdminSite())
        cls.user = UserFactory(is_superuser=True, is_staff=True)

    def test_contacts_count(self):
        # given
        sm = SyncManagerFactory()
        EveContactFactory(manager=sm)
        EveContactFactory(manager=sm)
        EveContactFactory(manager=sm)

        # when
        got = self.modeladmin._contacts_count(sm)

        # then
        self.assertEqual(got, "3")


class TestEveWarChangeList(TestCase):
    def test_can_open_page_normally(self):
        # given
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        SyncManagerFactory()

        # when
        response = self.client.get("/admin/standingssync/evewar/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)


class TestEvContactChangeList(TestCase):
    def test_can_open_page_normally(self):
        # given
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        sm = SyncManagerFactory()
        EveContactFactory(manager=sm)

        # when
        response = self.client.get("/admin/standingssync/evecontact/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
